# -*- coding: utf-8 -*-
"""
ITU-RPy is a python implementation of the ITU-P R Recommendations.

ITU-Rpy can be used to compute atmospheric attenuation for Earth-to-space
and horizontal paths, for frequencies in the GHz range.

The propagation loss on an Earth-space path and a horizontal-path, relative to
the free-space loss, is the sum of different contributions, namely:
 * attenuation by atmospheric gases;
 * attenuation by rain, other precipitation and clouds;
 * scintillation and multipath effects;
 * attenuation by sand and dust storms.

Each of these contributions has its own characteristics as a function of
frequency, geographic location and elevation angle. ITU-Rpy allows for fast,
vectorial computation of the different contributions to the atmospheric
attenuation.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__all__ = ['utils', 'plotting']
import itur.utils
import itur.plotting

from .models.itu618 import rain_attenuation, scintillation_attenuation
from .models.itu676 import gaseous_attenuation_slant_path, \
    gaseous_attenuation_inclined_path, \
    gaseous_attenuation_terrestrial_path
from .models.itu835 import standard_pressure
from .models.itu836 import surface_water_vapour_density, \
    total_water_vapour_content
from .models.itu840 import cloud_attenuation
from .models.itu1510 import surface_mean_temperature
from .models.itu1511 import topographic_altitude

import numpy as np
import warnings
import astropy.units as u

from .__version__ import __version__

AUTHORS = "Inigo del Portillo"


def atmospheric_attenuation_slant_path(
        lat, lon, f, el, p, D, hs=None, rho=None, R001=None, eta=0.5, T=None,
        H=None, P=None, hL=1e3, Ls=None, tau=45, V_t=None, mode='approx',
        return_contributions=False, include_rain=True, include_gas=True,
        include_scintillation=True, include_clouds=True):
    """
    Calculate long-term atmospheric attenuation statistics for slant paths.

    This function provides estimates of the long-term statistics of
    the slant-path atmospheric attenuation at a given location, for
    frequencies up to 55 GHz and percentages of time 0.001% < `p` < 50%.

    The model used is based on the guidelines provided in Section 2 of
    ITU-R P.618. If optional values are not provided they will be
    automatically computed using the procedures described in other ITU-R P.
    recommendations.

    Parameters
    ----------
    lat : number, sequence, or numpy.ndarray
        Latitudes of the receiver points
    lon : number, sequence, or numpy.ndarray
        Longitudes of the receiver points
    f : number or Quantity
        Frequency (GHz)
    el : sequence, number or Quantity
        Elevation angle (degrees)
    p : number
        Percentage of the time the rain attenuation value is exceeded.
    D: number or Quantity
        Physical diameter of the earth-station antenna (m)
    hs : number, sequence, or numpy.ndarray, optional
        Height above mean sea level of the earth station (km). If local data for
        the earth station height above mean sea level is not available, an
        estimate is obtained from the maps of topographic altitude
        given in Recommendation ITU-R P.1511.
    rho : number or Quantity, optional
        Water vapor density (g/m3). If not provided, an estimate is obtained
        from Recommendation Recommendation ITU-R P.836.
    R001: number  or Quantity, optional
        Point rainfall rate for the location for 0.01% of an average year \
        (mm/h). If not provided, an estimate is obtained from Recommendation
        ITU-R P.837. Some useful values:
            * 0.25 mm/h : Drizzle
            *  2.5 mm/h : Light rain
            * 12.5 mm/h : Medium rain
            * 25.0 mm/h : Heavy rain
            * 50.0 mm/h : Downpour
            * 100  mm/h : Tropical
            * 150  mm/h : Monsoon
    eta: number, optional
        Antenna efficiency. Default value 0.5 (conservative estimate)
    T: number, sequence, or numpy.ndarray, optional
        Average surface ambient temperature (°C) at the site. If None, uses the
        ITU-R P.1510 to estimate the surface ambient temperature.
    H: number, sequence, or numpy.ndarray, optional
        Average surface relative humidity (%) at the site. If None, uses the
        ITU-R P.836 to estimate the wet term of the surface relative humidity.
    P: number, sequence, or numpy.ndarray, optional
        Average surface pressure (hPa) at the site. If None, uses the
        ITU-R P.835 to estimate the average surface pressure.
    hL : number, optional
        Height of the turbulent layer (m). Default value 1000 m
    Ls :number, optional
        Slant path length (km). If not provided, it will be computed using the
        rain height and the elevation angle. The ITU model does not require
        this parameter as an input.
    tau : number, optional
        Polarization tilt angle relative to the horizontal (degrees)
        (tau = 45 deg for circular polarization). Default value is 45
    V_t : number or Quantity, optional
        Integrated water vapour content along the path (kg/m2 or mm).
        If not provided this value is estimated using Recommendation
        ITU-R P.836. Default value None
    mode : string, optional
        Mode for the calculation of gaseous attenuation. Valid values are
        'approx', 'exact'. If 'approx' Uses the method in Annex 2 of
        Recommendation ITU-R P.676, else uses the method described in
        Section 1. Default, 'approx'
    return_contributions: bool, optional
        Determines whether individual contributions from gases, rain, clouds
        and scintillation are returned in addition to the total attenuation
        (True), or just the total atmospheric attenuation (False).
        Default is False
    include_rain: bool, optional
        Determines whether to include the rain contribution in the total
        atmospheric attenuation calculation or not. Default is True
    include_gas: bool, optional
        Determines whether to include the gaseous contribution in the total
        atmospheric attenuation calculation or not. Default is True
    include_scintillation: bool, optional
        Determines whether to include the scintillation contribution in the
        total atmospheric attenuation calculation or not. Default is True
    include_clouds: bool, optional
        Determines whether to include the clouds contribution in the total
        atmospheric attenuation calculation or not. Default is True


    Returns
    -------
    A : Quantity
        Total atmospheric attenuation (dB)

    Ag, Ac, Ar, As, A : tuple
        Gaseous, Cloud, Rain, Scintillation contributions to total attenuation,
        and total attenuation (dB)



    References
    ----------
    [1] Propagation data and prediction methods required for the design of
    Earth-space telecommunication systems:
    https://www.itu.int/dms_pubrec/itu-r/rec/p/R-REC-P.618-12-201507-I!!PDF-E.pdf


    """
    if np.logical_or(p < 0.001, p > 50).any():
        warnings.warn(
            RuntimeWarning(
                'The method to compute the total '
                'atmospheric attenuation in recommendation ITU-P 618-13 '
                'is only recommended for unavailabilities (p) between '
                '0.001 % and 50 %'))

    # This takes account of the fact that a large part of the cloud attenuation
    # and gaseous attenuation is already included in the rain attenuation
    # prediction for time percentages below 1%. Eq. 64 and Eq. 65 in
    # Recommendation ITU 618-12
    p_c_g = np.maximum(1, p)

    # Estimate the ground station altitude
    if hs is None:
        hs = topographic_altitude(lat, lon)

    # Surface mean temperature
    if T is None:
        T = surface_mean_temperature(lat, lon)

    # Estimate the surface Pressure
    if P is None:
        P = standard_pressure(hs)

    # Estimate the surface Pressure
    if V_t is None:
        V_t = total_water_vapour_content(lat, lon, p_c_g, hs)

    # Estimate the surface water vapour density
    if rho is None:
        rho = surface_water_vapour_density(lat, lon, p_c_g, hs)

    # Compute the attenuation components
    if include_rain:
        Ar = rain_attenuation(lat, lon, f, el, hs, p, R001, tau, Ls)
    else:
        Ar = 0 * u.dB

    if include_gas:
        Ag = gaseous_attenuation_slant_path(f, el, rho, P, T, V_t, hs, mode)
    else:
        Ag = 0 * u.dB

    if include_clouds:
        Ac = cloud_attenuation(lat, lon, el, f, p_c_g)
    else:
        Ac = 0 * u.dB

    if include_scintillation:
        As = scintillation_attenuation(lat, lon, f, el, p, D, eta,
                                       T, H, P, hL)
    else:
        As = 0 * u.dB

    # Compute the total attenuation according to
    A = Ag + np.sqrt((Ar + Ac) ** 2 + As ** 2)

    if return_contributions:
        return Ag, Ac, Ar, As, A
    else:
        return A



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


import numpy as np
from astropy import units as u

from itur import utils
from itur.models.itu1144 import (bilinear_2D_interpolator,
                                 bicubic_2D_interpolator)
from itur.utils import (prepare_input_array, prepare_output_array,
                        load_data_interpolator)



class __ITU1510__():

    """Annual mean surface temperature

    Available versions include:
    * P.1510-0 (02/01) (Current version)
    """
    # This is an abstract class that contains an instance to a version of the
    # ITU-R P.1510 recommendation.

    def __init__(self, version=1):
        if version == 1:
            self.instance = _ITU1510_1_()
        elif version == 0:
            self.instance = _ITU1510_0_()
        else:
            raise ValueError('Version ' + str(version) + ' is not implemented'
                             ' for the ITU-R P.1510 model.')

    @property
    def __version__(self):
        return self.instance.__version__

    def surface_mean_temperature(self, lat, lon):
        """
        Method to compute the annual mean surface temperature (K).

        The temperature is computed at 2 m above the surface of the Earth.
        """
        return self.instance.temperature(lat, lon)

    def surface_month_mean_temperature(self, lat, lon, m):
        # Abstract method to compute the monthly surface mean temperature
        fcn = np.vectorize(self.instance.surface_month_mean_temperature,
                           excluded=[0, 1], otypes=[np.ndarray])
        return np.array(fcn(lat, lon, m).tolist())


class _ITU1510_1_():

    def __init__(self):
        self.__version__ = 1
        self.year = 2017
        self.month = 6
        self.link = 'https://www.itu.int/rec/R-REC-P.1510/' +\
                    'recommendation.asp?lang=en&parent=R-REC-P.1510-1-201706-I'

        self.__months = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]
        self._temperature = {}
        self._month_temperature = {}

    def temperature(self, lat, lon):
        if not self._temperature:
            self._temperature = load_data_interpolator(
                '1510/v1_lat.npz', '1510/v1_lon.npz',
                '1510/v1_t_annual.npz', bilinear_2D_interpolator)

        lon[lon > 180] = lon[lon > 180] - 360
        return self._temperature(
            np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    def month_temperature(self, lat, lon, m):
        if not self._month_temperature:
            for _m in self.__months:
                self._month_temperature[_m] = load_data_interpolator(
                    '1510/v1_lat.npz', '1510/v1_lon.npz',
                    '1510/v1_t_month{0:02d}.npz'.format(_m),
                    bilinear_2D_interpolator)

        lon[lon > 180] = lon[lon > 180] - 360
        return self._month_temperature[m](
            np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    def surface_mean_temperature(self, lat, lon):
        """
        Method to compute the annual mean surface temperature (K) at 2 m
        above the surface of the Earth
        """
        return self.temperature(lat, lon)

    def surface_month_mean_temperature(self, lat, lon, m):
        return self.month_temperature(lat, lon, m)


class _ITU1510_0_():

    def __init__(self):
        self.__version__ = 0
        self.year = 2001
        self.month = 2
        self.link = 'https://www.itu.int/rec/R-REC-P.1510/' +\
                    'recommendation.asp?lang=en&parent=R-REC-P.1510-0-200102-I'

        self._temperature = {}

    def temperature(self, lat, lon):
        if not self._temperature:
            self._temperature = load_data_interpolator(
                '1510/v0_lat.npz', '1510/v0_lon.npz',
                '1510/v0_temp.npz', bicubic_2D_interpolator)

        return self._temperature(
            np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    def surface_month_mean_temperature(self, lat, lon, m):
        raise NotImplementedError('The monthly mean temperature is not'
                                  'implemented in recomendation ITU-R P.1510'
                                  '-{0}'.format(self.__version__))


__model = __ITU1510__()


def change_version(new_version):
    """
    Change the version of the ITU-R P.1510 recommendation currently being used.


    This function changes the model used for the ITU-R P.1510 recommendation
    to a different version.

    Parameters
    ----------
    new_version : int
        Number of the version to use.
        Valid values are:

        * 1: Activates recommendation ITU-R P.1510-1 (06/17) (Current version)
        * 0: Activates recommendation ITU-R P.1510-0 (02/01) (Current version)
    """
    global __model
    __model = __ITU1510__(new_version)


def get_version():
    """
    Obtain the version of the ITU-R P.1510 recommendation currently being used.

    Returns
    -------
    version: int
        Version currently being used.
    """
    return __model.__version__


def surface_mean_temperature(lat, lon):
    """
    Annual mean surface temperature (K) at 2 m above the surface of the Earth.

    A method to estimate the annual mean surface temperature (K) at 2 m
    above the surface of the Earth


    Parameters
    ----------
    lat : number, sequence, or numpy.ndarray
        Latitudes of the receiver points
    lon : number, sequence, or numpy.ndarray
        Longitudes of the receiver points


    Returns
    -------
    annual_temperature: numpy.ndarray
        Annual mean surface temperature (K). Same dimensions as lat and lon.


    References
    ----------
    [1] Annual mean surface temperature:
    https://www.itu.int/rec/R-REC-P.1510/en

    """
    type_output = type(lat)
    lat = prepare_input_array(lat)
    lon = prepare_input_array(lon)
    lon = np.mod(lon, 360)
    val = __model.surface_mean_temperature(lat, lon)
    return prepare_output_array(val, type_output) * u.Kelvin


def surface_month_mean_temperature(lat, lon, m):
    """
    Monthly mean surface temperature (K) at 2 m above the surface of the Earth.

    A method to estimate the monthly mean surface temperature (K) at 2 m
    above the surface of the Earth


    Parameters
    ----------
    lat : number, sequence, or numpy.ndarray
        Latitudes of the receiver points
    lon : number, sequence, or numpy.ndarray
        Longitudes of the receiver points
    m : integer
      Index of the month (1=Jan, 2=Feb, 3=Mar, 4=Apr, ...)


    Returns
    -------
    monthly_temperature: numpy.ndarray
        Monthly mean surface temperature (K). Same dimensions as lat and lon.


    References
    ----------
    [1] Annual mean surface temperature:
    https://www.itu.int/rec/R-REC-P.1510/en

    """
    type_output = type(lat)
    lat = prepare_input_array(lat)
    lon = prepare_input_array(lon)
    lon = np.mod(lon, 360)
    val = __model.surface_month_mean_temperature(lat, lon, m)
    return prepare_output_array(val, type_output) * u.Kelvin



# -*- coding: utf-8 -*-
import sys
import numpy as np
import pandas as pd
import os.path as path
from collections import defaultdict

import unittest as test

import itur.models as models
from itur import atmospheric_attenuation_slant_path


pd.set_option('display.max_colwidth', -1)
basepath = path.dirname(path.realpath(__file__))
test_data = path.join(basepath, 'test_data')
html_path = path.join(basepath, '../docs/validation')


desc_validation =\
"""This page contains the validation examples for Recommendation {0}: {1}.

All test cases were extracted from the
`ITU Validation examples file (rev 5.1) <https://www.itu.int/en/ITU-R/study-groups/rsg3/ionotropospheric/CG-3M3J-13-ValEx-Rev5_1.xlsx>`_.

.. contents:: Functions tested
    :depth: 2

"""

desc_test_case =\
"""The table below contains the results of testing function ``{0}``.
The test cases were extracted from spreadsheet ``{1}`` from the
`ITU Validation examples file (rev 5.1) <https://www.itu.int/en/ITU-R/study-groups/rsg3/ionotropospheric/CG-3M3J-13-ValEx-Rev5_1.xlsx>`_.
In addition to the input-arguments, expected result (``ITU Validation``), and
ITU-Rpy computed result (``ITUR-py Result``), the absolute and relative errors
are shown. Each test case is color-coded depending on the magnitude of the
errors (green = pass, errors are negligible, red = fail, relative error is
above 0.01%).

In addition, the code snippet below shows an example of how to generate the
first row of the results in the table:

.. code-block:: python

{2}

"""

html_header = """
       <html>
         <head>
             <title>Validation results for {0}</title>
             <style>
                   table {{
                       border-collapse: collapse;
                       font-size: 10px;
                       width: 100%;
                       padding: 2px;
                     }}

                     th {{
                         background-color: black;
                         color: white;
                     }}

                     th, td {{
                       text-align: center;
                       padding: 2px;
                       width: 1%;
                       font-family: Arial, Helvetica, sans-serif;
                       white-space: nowrap;
                     }}

                     tr:nth-child(even) {{background-color: #f2f2f2;}}
                     tr:hover {{background-color: khaki;}}

               </style>
         </head>
         <body>
   """

html_footer = """
            <br><br>
          </body>
        </html>
        """

def create_ITU_suite():
    """ A test suite for the ITU-P Recommendations. Recommendations tested:
    * ITU-P R-676-11
    * ITU-P R-618-13
    * ITU-P R-453-14
    * ITU-P R-837-7
    * ITU-P R-838-3
    * ITU-P R-839-4
    * ITU-P R-840-8
    * ITU-P R-1511-1
    * ITU-P R-1511-2
    """
    suite = ITU_Suite()

    # ITU-R P.453 tests (Gaseous attenuation)
    suite.addTest(ITUR453_14TestCase('test_wet_term_radio_refractivity'))

    # ITU-R P.618
    suite.add_test(ITUR618_13TestCase('test_rain_attenuation'))
    suite.add_test(ITUR618_13TestCase('test_rain_probability'))
    suite.add_test(ITUR618_13TestCase('test_scintillation_attenuation'))
    suite.add_test(ITUR618_13TestCase('test_total_attenuation'))
    suite.add_test(ITUR618_13TestCase('test_cross_polarization_discrimination'))

    # ITU-R P.676
    suite.add_test(ITUR676_12TestCase('test_gamma0'))
    suite.add_test(ITUR676_12TestCase('test_gammaw'))
    suite.add_test(ITUR676_12TestCase('test_gamma'))
    suite.add_test(ITUR676_12TestCase('test_zenith_attenuation'))
    suite.add_test(ITUR676_12TestCase('test_attenuation_gas'))

    # ITU-R P.836
    suite.add_test(ITUR836_6TestCase('test_surface_water_vapour_density_annual'))
    suite.add_test(ITUR836_6TestCase('test_total_water_vapour_content_annual'))

    # ITU-R P.837
    suite.add_test(ITUR837_7TestCase('test_rainfall_rate'))
    suite.add_test(ITUR837_7TestCase('test_rainfall_rate_probability'))
    suite.add_test(ITUR837_7TestCase('test_rainfall_rate_R001'))

    # ITU-R P.838
    suite.add_test(ITUR838_3TestCase('test_rain_specific_attenuation'))

    # ITU-R P.839
    suite.add_test(ITUR839_4TestCase('test_isoterm_0_deg'))
    suite.add_test(ITUR839_4TestCase('test_rain_height'))

    # ITU-R P.840
    suite.add_test(ITUR840_8TestCase('test_columnar_content_reduced_liquid'))
    suite.add_test(ITUR840_8TestCase('test_cloud_attenuation'))

    # ITU-R P.1510
    suite.add_test(ITUR1510_1TestCase('test_surface_mean_temperature'))

    # ITU-R P.1511
    suite.add_test(ITUR1511_1TestCase('test_topographic_altitude'))
    suite.add_test(ITUR1511_2TestCase('test_topographic_altitude'))

    return suite

# Format HTML code

def formatter_fcn(s):
    return '\t\t\t<td style="text-align:left">' + str(s)


def formatter_rel_error_cell(s):
    if np.isnan(float(s)) or np.isinf(float(s)):
        return '\t\t\t<td bgcolor="cornflowerblue">{0:.3f}'.format(s)
    elif abs(float(s)) < 0.01:
        return '\t\t\t<td bgcolor="lightgreen">{0:.3f}'.format(s)
    else:
        return '\t\t\t<td bgcolor="salmon">{0:.3f}'.format(s)


def formatter_error(s):
    if np.isnan(float(s)):
        return '\t\t\t<td bgcolor="cornflowerblue">{0:.2e}'.format(s)
    elif abs(float(s)) < 0.1:
        return '\t\t\t<td bgcolor="lightgreen">{0:.2e}'.format(s)
    else:
        return '\t\t\t<td bgcolor="salmon">{0:.3e}'.format(s)


def format_table(table):
    # Fix cells with a cell within
    table = table.replace('<td><td', '<td')

    # Format headers
    table = table.replace('res_val', 'ITU Validation')
    table = table.replace('res_fcn', 'ITU-Rpy Result')
    table = table.replace('error_rel', 'Relative Error')
    table = table.replace('error', 'Absolute Error')
    table = table.replace('fcn', 'ITU-Rpy Function')

    return table


def formatter_digits(fcn, val):
    ret = []
    COL_STR = '<span style="color: {1}">{0}</span>'
    for f, v in zip(fcn, val):
        i_equal = 0

        # Convert numbers to strings
        s_f = '{:0.6f}'.format(f)
        s_v = '{:0.6f}'.format(v)

        # Determine how many numbers are equal
        for c_f, c_v in zip(s_f, s_v):
            if c_f == c_v:
                i_equal += 1
            else:
                break

        # Format the digits by coloring equal and different sections
        if i_equal > 0:
            s = COL_STR.format(s_f[:i_equal], 'darkgreen')
            if i_equal < len(s_f):
                s += COL_STR.format(s_f[i_equal:], 'darkred')
        else:
            s = COL_STR.format(s_f, 'darkred')

        ret.append(s)

    return ret


class ITU_Suite(test.TestSuite):

    def __init__(self):
        test.TestSuite.__init__(self)
        self.test_cases = {}

    def add_test(self, test_case):
        self.test_cases[test_case.__class__.__name__] = test_case
        self.addTest(test_case)

    def rst_reports(self, path_report=None):
        for test_name, test_case in self.test_cases.items():
            ret = test_case.produce_rst_report()
            if path_report:
                fpath = path.join(
                  path_report,
                  test_name.lower().replace('testcase', '') + '.rst')
                with open(fpath, 'w', encoding='utf-8') as fd:
                    fd.write(ret)


class ITU_TestCase(test.TestCase):
    report = defaultdict(dict)

    def read_csv(self, path_name, columns):
        self.path_name = path_name
        df = pd.read_csv(path_name, sep=',', skiprows=range(1, 2))
        units = pd.read_csv(path_name, sep=',', nrows=2, encoding='cp1252')
        self.units = dict(units[columns].iloc[0])
        return df[columns]

    def setUp(self):
        self.tests = []

    def __run__(self, test_name, test_fcn, df, attributes, result_value,
                n_places=5):

        test_fcn_name = test_fcn
        test_fcn = eval(test_fcn)

        # Add units for the result value and function
        self.units['res_val'] = self.units[result_value]
        self.units['res_fcn'] = self.units[result_value]
        # self.units['error'] = self.units[result_value]
        # self.units['error_rel'] = '(%)'

        # Evaluate all the functions
        res = []
        for i, row in df.iterrows():
            args = {a: row[a] for a in attributes}
            # Evaluate function
            res_fcn = test_fcn(**args)
            res_val = row[result_value]

            # Format dictionary to be added to report
            line = dict(args)
            line['fcn'] = test_fcn_name
            line['res_fcn'] = res_fcn.value
            line['res_val'] = res_val
            line['error'] = (res_val - res_fcn.value)
            if res_val == 0 and abs(line['error']) < 1e-6:
                line['error_rel'] = 0
            else:
                line['error_rel'] = round(
                      (res_val - res_fcn.value) / res_val * 100, 3)

            res.append(line)

        # Create data frame with the report
        order = ['fcn'] + attributes + ['res_val', 'res_fcn', 'error',
                                        'error_rel']
        df = pd.DataFrame(res)
        self.report[self.__class__.__name__][test_name] = {
            'df': df,
            'class_name': self.__class__.__name__,
            'test_name': test_name,
            'units': self.units,
            'path_csv': self.path_name,
            'test_fcn': test_fcn_name,
            'attributes': attributes,
            'n_places': n_places,
            'report_html': df[order]
          }

        # Do the assert equal for all the tests
        for ret in res:
            res_val = ret['res_val']
            res_fcn = ret['res_fcn']
            try:
                self.assertAlmostEqual(res_val, res_fcn, places=n_places)
            except AssertionError as e:
                print(e)

    def generate_code_example(self, report):
        """Generate a code example of the call used for this function

        Parameters
        ----------
        report : dict
          The dictionary containing the parameters used in a test case.

        Returns
        -------
        str
          A string with an example of code used to call this test.
        """
        ret = ["    import itur", "", "    # Define input attributes"]
        attributes = report['attributes']
        test_fcn = report['test_fcn']
        test_fcn_name = report['test_fcn'].split('.')[-1]

        row_1 = report['df'].iloc[0]
        units = report['units']

        # Write the attributesf
        for attr_name, attr_val in zip(attributes, row_1[attributes]):
            ret.append('    {0} = {1}  # {2}'.format(
                attr_name, attr_val, units[attr_name]))

        # Add call to test-function
        ret.extend([
            "", "    # Make call to test-function {0}".format(test_fcn_name),
            '    itur_val = itur.{0}({1})'.format(
              test_fcn, ", ".join([att + '=' + att for att in attributes]))])

        # Compute errors
        ret.extend([
            "",
            "    # Compute error with respect to value in ITU example file",
            '    ITU_example_val = {0}  # {1}'.format(
                        row_1['res_val'], units['res_val']),
            '    error = ITU_example_val - itur_val.value',
            '    error_rel = error / ITU_example_val * 100  # (%)'])

        return "\n".join(ret)

    def produce_rst_report(self):

        ret = []
        title = "Validation results for {0}".format(self.itu_name)
        ret.append(title)
        ret.append('=' * len(title))
        ret.append('')
        ret.append(desc_validation.format(self.itu_name, self.itu_description))

        for test_name in self.report[self.__class__.__name__]:

            # Create HTML table for this test
            report = self.report[self.__class__.__name__][test_name]
            table = self.create_html_table(report, include_header=True)
            html_file = path.join(html_path, test_name + '_table.html')
            with open(html_file, 'w', encoding='utf-8') as fd:
                fd.write(table)

            # Create HTML table for this test
            test_fcn_name = report['test_fcn'].split('.')[-1]
            test_case_name = "Function {0}".format(test_fcn_name)
            ret.append(test_case_name)
            ret.append("-" * len(test_case_name))
            ret.append("")
            ret.append(desc_test_case.format(
                    test_fcn_name, path.basename(report['path_csv']),
                    self.generate_code_example(report)))

            ret.extend(['.. raw:: html',
                        '    :file: {0}'.format(test_name + '_table.html'),
                        '', ''])

        return "\n".join(ret)

    def create_html_table(self, report, include_header=False):
        fmtrs = {'error_rel': formatter_rel_error_cell,
                 'error': formatter_error,
                 'fcn': formatter_fcn}

        if include_header:
            table = html_header.format(report['test_fcn'])

        df = report['report_html']
        df['res_fcn'] = formatter_digits(df['res_fcn'], df['res_val'])

        # Add units to header attributes
        col_dict = {}
        for col in df.columns:
            if col in report['units']:
                col_dict[col] = '{0} {1}'.format(col, report['units'][col])
            else:
                col_dict[col] = col

        df.rename(columns=col_dict, inplace=True)
        table += df.to_html(bold_rows=True, index=False, justify='center',
                            table_id=report['test_name'].lower(), escape=False,
                            formatters=fmtrs)

        table = format_table(table)
        if include_header:
            table += html_footer
        return table


class ITUR453_14TestCase(ITU_TestCase):

    itu_name = 'ITU-R P.453-14'
    itu_description = 'TBD'

    def test_wet_term_radio_refractivity(self):
        # Set the version to the
        models.itu453.change_version(13)

        # Read the test data
        df = self.read_csv(path.join(test_data, '453/ITURP453-14_Nwet.csv'),
                           columns=['lat', 'lon', 'p', 'Nwet'])

        # Run test and generate the report
        self.__run__('test_wet_term_radio_refractivity',
                     test_fcn='models.itu453.map_wet_term_radio_refractivity',
                     df=df, attributes=['lat', 'lon', 'p'],
                     result_value='Nwet',
                     n_places=5)


class ITUR618_13TestCase(ITU_TestCase):

    itu_name = 'ITU-R P.618-13'
    itu_description = 'Propagation data and prediction methods required for' +\
                      ' the design of Earth-space telecommunication systems'

    def test_rain_attenuation(self):
        # Set the version to the
        models.itu618.change_version(13)

        # Read the test data
        df = self.read_csv(path.join(test_data, '618/ITURP618-13_A_rain.csv'),
                           columns=['lat', 'lon', 'hs', 'el', 'f', 'tau', 'p',
                                    'R001', 'A_rain'])

        # Run test and generate the report
        self.__run__('test_rain_attenuation',
                     test_fcn='models.itu618.rain_attenuation',
                     df=df, attributes=['lat', 'lon', 'hs', 'el', 'f',
                                        'tau', 'p', 'R001'],
                     result_value='A_rain',
                     n_places=5)

    def test_rain_probability(self):
        # Set the version to the
        models.itu618.change_version(13)

        # Read the test data
        df = self.read_csv(path.join(test_data, '618/ITURP618-13_A_rain.csv'),
                           columns=['lat', 'lon', 'hs', 'el', 'Ls', 'P0',
                                    'P_rain'])

        # Run test and generate the report
        self.__run__('test_rain_probability',
                     test_fcn='models.itu618.rain_attenuation_probability',
                     df=df, attributes=['lat', 'lon', 'hs', 'el', 'Ls', 'P0'],
                     result_value='P_rain',
                     n_places=5)

    def test_scintillation_attenuation(self):
        # Set the version to the
        models.itu618.change_version(13)

        # Read the test data
        df = self.read_csv(path.join(test_data, '618/ITURP618-13_A_rain.csv'),
                           columns=['lat', 'lon', 'f', 'el', 'p', 'D', 'eta',
                                    'A_scin'])

        # Run test and generate the report
        self.__run__('test_scintillation_attenuation',
                     test_fcn='models.itu618.scintillation_attenuation',
                     df=df, attributes=['lat', 'lon', 'f', 'el', 'p', 'D',
                                        'eta'],
                     result_value='A_scin',
                     n_places=5)

    def test_cross_polarization_discrimination(self):
        # Set the version to the
        models.itu618.change_version(13)

        # Read the test data
        df = self.read_csv(path.join(test_data, '618/ITURP618-13_A_xpd.csv'),
                           columns=['f', 'el', 'p', 'tau', 'Ap', 'XPD'])

        # Run test and generate the report
        self.__run__('test_cross_polarization_discrimination',
                     test_fcn='models.itu618.rain_cross_polarization_discrimination',
                     df=df, attributes=['f', 'el', 'p', 'tau', 'Ap'],
                     result_value='XPD',
                     n_places=5)

    def test_total_attenuation(self):
        # Set the version to the
        models.itu618.change_version(13)

        # Read the test data
        df = self.read_csv(path.join(test_data, '618/ITURP618-13_A_total.csv'),
                           columns=['lat', 'lon', 'f', 'el', 'p', 'D', 'eta',
                                    'tau', 'hs', 'A_total'])

        # Run test and generate the report
        self.__run__('test_total_attenuation',
                     test_fcn='atmospheric_attenuation_slant_path',
                     df=df, attributes=['lat', 'lon', 'f', 'el', 'p', 'D',
                                        'eta', 'tau', 'hs'],
                     result_value='A_total',
                     n_places=4)


class ITUR676_12TestCase(ITU_TestCase):

    itu_name = 'ITU-R P.676-12'
    itu_description = 'Attenuation by atmospheric gases and related effects'

    def test_gamma0(self):
        # Set the version to the
        models.itu676.change_version(12)

        path_file = '676/ITURP676-12_gamma.csv'
        # Read the test data
        df = self.read_csv(path.join(test_data, path_file),
                           columns=['f', 'P', 'rho', 'T', 'gamma0'])

        # Run test and generate the report
        self.__run__('test_gamma0',
                     test_fcn='models.itu676.gamma0_exact',
                     df=df, attributes=['f', 'P', 'rho', 'T'],
                     result_value='gamma0',
                     n_places=5)

    def test_gammaw(self):
        # Set the version to the
        models.itu676.change_version(12)

        path_file = '676/ITURP676-12_gamma.csv'
        # Read the test data
        df = self.read_csv(path.join(test_data, path_file),
                           columns=['f', 'P', 'rho', 'T', 'gammaw'])

        # Run test and generate the report
        self.__run__('test_gammaw',
                     test_fcn='models.itu676.gammaw_exact',
                     df=df, attributes=['f', 'P', 'rho', 'T'],
                     result_value='gammaw',
                     n_places=5)

    def test_gamma(self):
        # Set the version to the
        models.itu676.change_version(12)

        path_file = '676/ITURP676-12_gamma.csv'
        # Read the test data
        df = self.read_csv(path.join(test_data, path_file),
                           columns=['f', 'P', 'rho', 'T', 'gamma'])

        # Run test and generate the report
        self.__run__('test_gamma',
                     test_fcn='models.itu676.gamma_exact',
                     df=df, attributes=['f', 'P', 'rho', 'T'],
                     result_value='gamma',
                     n_places=5)

    def test_attenuation_gas(self):
        # Set the version to the
        models.itu676.change_version(12)

        path_file = '676/ITURP676-12_A_gas.csv'
        # Read the test data
        df = self.read_csv(path.join(test_data, path_file),
                           columns=['f', 'el', 'P', 'rho', 'T', 'h', 'V_t',
                                    'A_gas'])

        # Run test and generate the report
        self.__run__('test_attenuation_gas',
                     test_fcn='models.itu676.gaseous_attenuation_slant_path',
                     df=df, attributes=['f', 'el', 'rho', 'P', 'T', 'h',
                                        'V_t'],
                     result_value='A_gas',
                     n_places=5)

    def test_zenith_attenuation(self):
        # Set the version to the
        models.itu676.change_version(12)

        path_file = '676/ITURP676-12_zenith_attenuation.csv'
        # Read the test data
        df = self.read_csv(path.join(test_data, path_file),
                           columns=['lat', 'lon', 'p', 'f', 'h', 'V_t', 'Aw'])

        # Run test and generate the report
        self.__run__('test_zenith_attenuation',
                     test_fcn='models.itu676.zenit_water_vapour_attenuation',
                     df=df, attributes=['lat', 'lon', 'p', 'f', 'h', 'V_t'],
                     result_value='Aw',
                     n_places=5)


class ITUR836_6TestCase(ITU_TestCase):

    itu_name = 'ITU-R P.836-6'
    itu_description = 'Water vapour: surface density and total columnar content'

    def test_surface_water_vapour_density_annual(self):
        # Set the version to the
        models.itu836.change_version(6)

        path_file = '836/ITURP836-6_surface_water_vapour_density_annual.csv'
        # Read the test data
        df = self.read_csv(path.join(test_data, path_file),
                           columns=['lat', 'lon', 'alt', 'p', 'rho'])

        # Run test and generate the report
        self.__run__('test_surface_water_vapour_density_annual',
                     test_fcn='models.itu836.surface_water_vapour_density',
                     df=df, attributes=['lat', 'lon', 'alt', 'p'],
                     result_value='rho',
                     n_places=5)

    def test_total_water_vapour_content_annual(self):
        # Set the version to the
        models.itu836.change_version(6)

        path_file = '836/ITURP836-6_total_water_vapour_content_annual.csv'
        # Read the test data
        df = self.read_csv(path.join(test_data, path_file),
                           columns=['lat', 'lon', 'alt', 'p', 'V'])

        # Run test and generate the report
        self.__run__('test_total_water_vapour_content_annual',
                     test_fcn='models.itu836.total_water_vapour_content',
                     df=df, attributes=['lat', 'lon', 'alt', 'p'],
                     result_value='V',
                     n_places=5)


class ITUR837_7TestCase(ITU_TestCase):

    itu_name = 'ITU-R P.837-7'
    itu_description = 'Characteristics of precipitation for propagation ' +\
                      'modelling'

    def test_rainfall_rate(self):
        # Set the version to the
        models.itu837.change_version(7)

        path_file = '837/ITURP837-7_rainfall_rate.csv'
        # Read the test data
        df = self.read_csv(path.join(test_data, path_file),
                           columns=['lat', 'lon', 'p', 'Rp'])

        # Run test and generate the report
        self.__run__('test_rainfall_rate',
                     test_fcn='models.itu837.rainfall_rate',
                     df=df, attributes=['lat', 'lon', 'p'],
                     result_value='Rp',
                     n_places=3)

    def test_rainfall_rate_R001(self):
        # Set the version to the
        models.itu837.change_version(7)

        path_file = '837/ITURP837-7_rainfall_rate_R001.csv'
        # Read the test data
        df = self.read_csv(path.join(test_data, path_file),
                           columns=['lat', 'lon', 'p', 'Rp'])

        # Run test and generate the report
        self.__run__('test_rainfall_rate_R001',
                     test_fcn='models.itu837.rainfall_rate',
                     df=df, attributes=['lat', 'lon', 'p'],
                     result_value='Rp',
                     n_places=5)

    def test_rainfall_rate_probability(self):
        # Set the version to the
        models.itu837.change_version(7)

        path_file = '837/ITURP837-7_rainfall_rate_probability.csv'
        # Read the test data
        df = self.read_csv(path.join(test_data, path_file),
                           columns=['lat', 'lon', 'p'])

        # Run test and generate the report
        self.__run__('test_rainfall_rate_probability',
                     test_fcn='models.itu837.rainfall_probability',
                     df=df, attributes=['lat', 'lon'],
                     result_value='p',
                     n_places=5)


class ITUR838_3TestCase(ITU_TestCase):

    itu_name = 'ITU-R P.838-3'
    itu_description = 'Specific attenuation model for rain for use in ' +\
                      'prediction methods'

    def test_rain_specific_attenuation(self):
        # Set the version to the
        models.itu838.change_version(3)

        path_file = '838/ITURP838-3_rain_specific_attenuation.csv'
        # Read the test data
        df = self.read_csv(path.join(test_data, path_file),
                           columns=['el', 'f', 'R', 'tau', 'gamma_r'])

        # Run test and generate the report
        self.__run__('test_rain_specific_attenuation',
                     test_fcn='models.itu838.rain_specific_attenuation',
                     df=df, attributes=['el', 'f', 'R', 'tau'],
                     result_value='gamma_r',
                     n_places=5)


class ITUR839_4TestCase(ITU_TestCase):

    itu_name = 'ITU-R P.839-4'
    itu_description = 'Rain height model for prediction methods'

    def test_isoterm_0_deg(self):
        # Set the version to the
        models.itu839.change_version(4)

        path_file = '839/ITURP839-4_rain_height.csv'
        # Read the test data
        df = self.read_csv(path.join(test_data, path_file),
                           columns=['lat', 'lon', 'h0'])

        # Run test and generate the report
        self.__run__('test_isoterm_0_deg',
                     test_fcn='models.itu839.isoterm_0',
                     df=df, attributes=['lat', 'lon'],
                     result_value='h0',
                     n_places=5)

    def test_rain_height(self):
        # Set the version to the
        models.itu839.change_version(4)

        path_file = '839/ITURP839-4_rain_height.csv'
        # Read the test data
        df = self.read_csv(path.join(test_data, path_file),
                           columns=['lat', 'lon', 'hr'])

        # Run test and generate the report
        self.__run__('test_rain_height',
                     test_fcn='models.itu839.rain_height',
                     df=df, attributes=['lat', 'lon'],
                     result_value='hr',
                     n_places=5)


class ITUR840_8TestCase(ITU_TestCase):

    itu_name = 'ITU-R P.840-8'
    itu_description = 'Attenuation due to clouds and fog'

    def test_columnar_content_reduced_liquid(self):
        # Set the version to the
        models.itu840.change_version(8)

        path_file = '840/ITURP840-8_columnar_content_reduced_liquid.csv'
        # Read the test data
        df = self.read_csv(path.join(test_data, path_file),
                           columns=['lat', 'lon', 'p', 'Lred'])

        # Run test and generate the report
        self.__run__('test_columnar_content_reduced_liquid',
                     test_fcn='models.itu840.columnar_content_reduced_liquid',
                     df=df, attributes=['lat', 'lon', 'p'],
                     result_value='Lred',
                     n_places=5)

    def test_cloud_attenuation(self):
        # Set the version to the
        models.itu840.change_version(8)

        path_file = '840/ITURP840-8_cloud_attenuation.csv'
        # Read the test data
        df = self.read_csv(path.join(test_data, path_file),
                           columns=['lat', 'lon', 'f', 'el', 'p', 'Ac'])

        # Run test and generate the report
        self.__run__('test_cloud_attenuation',
                     test_fcn='models.itu840.cloud_attenuation',
                     df=df, attributes=['lat', 'lon', 'f', 'el', 'p'],
                     result_value='Ac',
                     n_places=5)


class ITUR1510_1TestCase(ITU_TestCase):

    itu_name = 'ITU-R P.1510-1'
    itu_description = 'Mean surface temperature'

    def test_surface_mean_temperature(self):
        # Set the version to the
        models.itu1510.change_version(1)

        path_file = '1510/ITURP1510-1_temperature.csv'
        # Read the test data
        df = self.read_csv(path.join(test_data, path_file),
                           columns=['lat', 'lon', 'T'])

        # Run test and generate the report
        self.__run__('test_surface_mean_temperature',
                     test_fcn='models.itu1510.surface_mean_temperature',
                     df=df, attributes=['lat', 'lon'],
                     result_value='T',
                     n_places=5)


class ITUR1511_1TestCase(ITU_TestCase):

    itu_name = 'ITU-R P.1511-1'
    itu_description = 'Topography for Earth-to-space propagation modelling'

    def test_topographic_altitude(self):
        # Set the version to the
        models.itu1511.change_version(1)

        path_file = '1511/ITURP1511-1_topographic_altitude.csv'
        # Read the test data
        df = self.read_csv(path.join(test_data, path_file),
                           columns=['lat', 'lon', 'hs'])

        # Run test and generate the report
        self.__run__('test_topographic_altitude',
                     test_fcn='models.itu1511.topographic_altitude',
                     df=df, attributes=['lat', 'lon'],
                     result_value='hs',
                     n_places=5)


class ITUR1511_2TestCase(ITU_TestCase):

    itu_name = 'ITU-R P.1511-2'
    itu_description = 'Topography for Earth-to-space propagation modelling'

    def test_topographic_altitude(self):
        # Set the version to the
        models.itu1511.change_version(2)

        path_file = '1511/ITURP1511-2_topographic_altitude.csv'
        # Read the test data
        df = self.read_csv(path.join(test_data, path_file),
                           columns=['lat', 'lon', 'hs'])

        # Run test and generate the report
        self.__run__('test_topographic_altitude',
                     test_fcn='models.itu1511.topographic_altitude',
                     df=df, attributes=['lat', 'lon'],
                     result_value='hs',
                     n_places=5)


if __name__ == '__main__':
    suite = create_ITU_suite()
    print('Validation tests for the ITU-R models')
    print('------------------------')
    print('A total of %d test-cases are going to be tested' %
          suite.countTestCases())
    sys.stdout.flush()
    test.TextTestRunner(verbosity=2).run(suite)
    # suite.html_reports(html_path)
    suite.rst_reports(html_path)



# -*- coding: utf-8 -*-
"""``itur.plotting`` provides convenient function to plot maps in ITU-Rpy.

This submodule uses ``matplotlib`` and ``cartopy`` as the default library to
plot maps. Alternatively, the user can use ``basemap`` (if installed).

The example below shows the use of ``plot_in_map`` to display the mean surface
temperature on the Earth.

.. code-block:: python

    import itur

    # Generate a regular grid of latitude and longitudes with 0.1 degree
    #  resolution.
    lat, lon = itur.utils.regular_lat_lon_grid(resolution_lat=0.1,
                                               resolution_lon=0.1)

    # Compute the surface mean temperature
    T = itur.models.itu1510.surface_mean_temperature(lat, lon)

    # Display the results in a map (using cartopy)
    ax = itur.plotting.plot_in_map(
            T, lat, lon, cmap='jet', vmin=230, vmax=310,
            cbar_text='Annual mean surface temperature [K]')

    # Display the results in a map (using basemap)
    ax = itur.plotting.plot_in_map_basemap(
            T, lat, lon, cmap='jet', vmin=230, vmax=310,
            cbar_text='Annual mean surface temperature [K]')

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np

try:
    import cartopy.crs as ccrs
    import cartopy.feature as cpf
    import cartopy.mpl.gridliner as cmpl
    plotting_installed = True
except BaseException:
    plotting_installed = False


def plot_in_map(data, lat=None, lon=None, lat_min=None, lat_max=None,
                lon_min=None, lon_max=None, cbar_text='', ax=None,
                figsize=(6, 4), **kwargs):
    """Plot the values in `data` in a map using ``cartopy``.

    The map uses an PlateCarree projection. Either
    {``lat``, ``lon``} or {``lat_min``, ``lat_max``, ``lon_min``, ``lon_max``}
    need to be provided as inputs. This function requires that ``cartopy``
    and ``matplotlib`` are installed.

    Parameters
    ----------
    data : np.ndarray
        Data values to be plotted.
    lat : np.ndarray
        Matrix with the latitudes for each point in data (deg N)
    lon : np.ndarray
        Matrix with the longitudes for each point in data (deg E)
    lat_min :  float
        Minimum latitude of the data (deg N)
    lat_max :  float
        Maximum latitude of the data (deg N)
    lon_min :  float
        Minimum longitude of the data (deg E)
    lat_max :  float
        Maximum longitude of the data (deg E)
    cbar_text : string
        Colorbar text caption.
    ax : Axes
        matplotlib axes where the data will be plotted.
    figsize : tuple
        Dimensions of the Figure
    **kwargs: dict
        Key-value arguments that will be passed to the contourf function.

    Returns
    -------
    ax : Axes
        The matploltib axes object
    """
    import matplotlib.pyplot as plt

    if not plotting_installed:
        raise RuntimeError('Neither cartopy nor matplotlib are installed. '
                           'Therefore plot_in_map cannot be used. '
                           'To use this function you need to install '
                           'the cartopy and matplotlib libraries')

    if all([el is None for el in [lat, lon, lat_min, lon_min,
                                  lat_max, lon_max]]):
        raise ValueError('Either {{lat, lon}} or {{lat_min, lon_min, lat_max,'
                         'lon_max}} need to be provided')

    elif lat is not None and lon is not None:
        if not(np.shape(lat) == np.shape(lon) and
               np.shape(lat) == np.shape(data)):
            raise RuntimeError('Shape of latitude grid is not equal to shape'
                               'of longitude grid')
        lat_max = np.max(lat)
        lat_min = np.min(lat)
        lon_max = np.max(lon)
        lon_min = np.min(lon)

    if ax is None:
        fig = plt.figure(figsize=figsize)
        proj = ccrs.PlateCarree(central_longitude=0.0)
        ax = fig.add_subplot(111, projection=proj)

    ax.set_extent([lon_min, lon_max, lat_min, lat_max],
                  crs=ccrs.PlateCarree())
    ax.coastlines(color='grey', linewidth=0.8)
    ax.add_feature(cpf.BORDERS, edgecolor='grey')

    parallels = np.arange(-80, 81, 20)
    meridians = np.arange(-180., 181., 30.)

    ax.gridlines(xlocs=meridians, ylocs=parallels, draw_labels=True,
                 color='white', linestyle=':', linewidth=0.2)

    im = ax.contourf(lon, lat, data, 100, transform=ccrs.PlateCarree(),
                     **kwargs)

    cbar = fig.colorbar(im, orientation='horizontal', fraction=0.046, pad=0.04)
    cbar.set_label(cbar_text)
    fig.show()
    return ax


def plot_in_map_basemap(data, lat=None, lon=None, lat_min=None,
                        lat_max=None, lon_min=None, lon_max=None,
                        cbar_text='', ax=None, figsize=(6, 4), **kwargs):
    """Plot the values in `data` in a map using ``basemap``.

    The map uses an equidistant cylindrical projection. Either
    {``lat``, ``lon``} or {``lat_min``, ``lat_max``, ``lon_min``, ``lon_max``}
    to be provided as inputs. This function requires that ``basemap`` and
    ``matplotlib`` are installed.

    Parameters
    ----------
    data : np.ndarray
        Data values to be plotted.
    lat : np.ndarray
        Matrix with the latitudes for each point in data (deg N)
    lon : np.ndarray
        Matrix with the longitudes for each point in data (deg E)
    lat_min :  float
        Minimum latitude of the data (deg N)
    lat_max :  float
        Maximum latitude of the data (deg N)
    lon_min :  float
        Minimum longitude of the data (deg E)
    lat_max :  float
        Maximum longitude of the data (deg E)
    cbar_text : string
        Colorbar text caption.
    ax : Axes
        matplotlib axes where the data will be plotted.
    figsize : tuple
        Dimensions of the Figure
    **kwargs: dict
        Key-value arguments that will be passed to the imshow function.

    Returns
    -------
    m : Basemap
        The map object generated by Basemap
    """
    try:
        import matplotlib.pyplot as plt
        from mpl_toolkits.basemap import Basemap
    except BaseException:
        raise RuntimeError('Basemap is not installed and therefore '
                           'plot_in_map_basemap cannot be used. To use this '
                           'function you need to install the basemap library')

    if all([el is None for el in [lat, lon, lat_min, lon_min,
                                  lat_max, lon_max]]):
        raise ValueError('Either {{lat, lon}} or {{lat_min, lon_min, lat_max,'
                         'lon_max}} need to be provided')

    elif lat is not None and lon is not None:
        if not(np.shape(lat) == np.shape(lon) and
               np.shape(lat) == np.shape(data)):
            raise RuntimeError('Shape of latitude grid is not equal to shape'
                               'of longitude grid')
        lat_max = np.max(lat)
        lat_min = np.min(lat)
        lon_max = np.max(lon)
        lon_min = np.min(lon)

    if ax is None:
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(111)

    m = Basemap(ax=ax, projection='cyl', llcrnrlat=lat_min,
                urcrnrlat=lat_max, llcrnrlon=lon_min, urcrnrlon=lon_max,
                resolution='l')

    m.drawcoastlines(color='grey', linewidth=0.8)
    m.drawcountries(color='grey', linewidth=0.8)
    parallels = np.arange(-80, 81, 20)
    m.drawparallels(parallels, labels=[1, 0, 0, 1], dashes=[2, 1],
                    linewidth=0.2, color='white')
    meridians = np.arange(0., 360., 30.)
    m.drawmeridians(meridians, labels=[1, 0, 0, 1], dashes=[2, 1],
                    linewidth=0.2, color='white')

    im = m.imshow(np.flipud(data), **kwargs)
    cbar = m.colorbar(im, location='bottom', pad="8%")
    cbar.set_label(cbar_text)
    return m



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


import numpy as np
from astropy import units as u

from itur import utils
from itur.models.itu1144 import bicubic_2D_interpolator
from itur.utils import (prepare_input_array, prepare_output_array,
                        load_data_interpolator)


class __ITU1511():
    """Topography for Earth-to-space propagation modelling. This model shall be
    used to obtain the height above mean sea level when no local data are
    available or when no data with a better spatial resolution is available.

    Available versions include:
    * P.1511-0 (02/01) (Superseded)
    * P.1511-1 (07/15) (Superseded)
    * P.1511-2 (08/19) (Current version)
    """
    # This is an abstract class that contains an instance to a version of the
    # ITU-R P.1511 recommendation.

    def __init__(self, version=2):
        if version == 2:
            self.instance = _ITU1511_2_()
        elif version == 1:
            self.instance = _ITU1511_1_()
        elif version == 0:
            self.instance = _ITU1511_0_()
        else:
            raise ValueError('Version ' + str(version) + ' is not implemented'
                             ' for the ITU-R P.1511 model.')

    @property
    def __version__(self):
        """
        Version of the model (similar to version of the ITU Recommendation)
        """
        return self.instance.__version__

    def topographic_altitude(self, lat, lon):
        # Abstract method to compute the topographic altitude
        return self.instance.topographic_altitude(lat, lon)


class _ITU1511_2_():
    """
    The values of topographical height (km) above mean sea level of the surface
    of the Earth are  provided on a 0.5° grid in both latitude and longitude.
    For a location different from the gridpoints, the height above mean sea
    level at the desired location can be obtained by performing a bi-cubic
    interpolation.
    """

    def __init__(self):
        self.__version__ = 2
        self.year = 2019
        self.month = 8
        self.link = 'https://www.itu.int/rec/R-REC-P.1511/' +\
                    'recommendation.asp?lang=en&parent=R-REC-P.1511-2-201908-I'

        self._altitude = None
        self._wgs4_altitude = None

    def altitude(self, lat, lon):
        if not self._altitude:
            self._altitude = load_data_interpolator(
                '1511/v2_lat.npz', '1511/v2_lon.npz',
                '1511/v2_topo.npz', bicubic_2D_interpolator)

        return self._altitude(
                np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    def wgs4_altitude(self, lat, lon):
        if not self._wgs4_altitude:
            self._wgs4_altitude = load_data_interpolator(
                '1511/v2_lat.npz', '1511/v2_lon.npz',
                '1511/v2_egm2008.npz', bicubic_2D_interpolator)

        return self._wgs4_altitude(
                np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    def topographic_altitude(self, lat_d, lon_d):
        """
        Method to compute the values of topographical height (km) above mean
        sea level of the surface of the Earth.
        """

        # The new recommendation provides the output in meters and uses
        # a -180 to 180 longitude refernce
        lon_d = np.where(lon_d > 180, lon_d - 360, lon_d)
        return self.altitude(lat_d, lon_d) / 1000


class _ITU1511_1_():
    """
    The values of topographical height (km) above mean sea level of the surface
    of the Earth are  provided on a 0.5° grid in both latitude and longitude.
    For a location different from the gridpoints, the height above mean sea
    level at the desired location can be obtained by performing a bi-cubic
    interpolation.
    """

    def __init__(self):
        self.__version__ = 1
        self.year = 2015
        self.month = 7
        self.link = 'https://www.itu.int/rec/R-REC-P.1511/' +\
                    'recommendation.asp?lang=en&parent=R-REC-P.1511-1-201507-I'

        self._altitude = None

    def altitude(self, lat, lon):
        if not self._altitude:
            self._altitude = load_data_interpolator(
                '1511/v1_lat.npz', '1511/v1_lon.npz',
                '1511/v1_topo_0dot5.npz', bicubic_2D_interpolator)

        return self._altitude(
                np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    def topographic_altitude(self, lat_d, lon_d):
        """
        Method to compute the values of topographical height (km) above mean
        sea level of the surface of the Earth.
        """
        return self.altitude(lat_d, lon_d)


class _ITU1511_0_():
    """
    The values of topographical height (km) above mean sea level of the surface
    of the Earth are  provided on a 0.5° grid in both latitude and longitude.
    For a location different from the gridpoints, the height above mean sea
    level at the desired location can be obtained by performing a bi-cubic
    interpolation.
    """

    def __init__(self):
        self.__version__ = 0
        self.year = 2001
        self.month = 2
        self.link = 'https://www.itu.int/rec/R-REC-P.1511/' +\
                    'recommendation.asp?lang=en&parent=R-REC-P.1511-0-200102-I'

        self._altitude = None

    def altitude(self, lat, lon):
        if not self._altitude:
            self._altitude = load_data_interpolator(
                '1511/v1_lat.npz', '1511/v1_lon.npz',
                '1511/v1_topo_0dot5.npz', bicubic_2D_interpolator)

        return self._altitude(
                np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    def topographic_altitude(self, lat_d, lon_d):
        """
        Method to compute the values of topographical height (km) above mean
        sea level of the surface of the Earth.
        """
        return self.altitude(lat_d, lon_d)


__model = __ITU1511()


def change_version(new_version):
    """
    Change the version of the ITU-R P.1511 recommendation currently being used.


    Parameters
    ----------
    new_version : int
        Number of the version to use.
        Valid values are:
          * 1:  Activates recommendation ITU-R P.1511-1 (07/15) (Current version)
          * 0:  Activates recommendation ITU-R P.1511-0 (02/01) (Superseded)

    """
    global __model
    __model = __ITU1511(new_version)


def get_version():
    """
    Obtain the version of the ITU-R P.1511 recommendation currently being used.

    Returns
    -------
    version: int
        Version currently being used.
    """
    return __model.__version__


def topographic_altitude(lat, lon):
    """
    Topographical height (km) above mean sea level of the surface of the Earth.


    Parameters
    ----------
    lat : number, sequence, or numpy.ndarray
        Latitudes of the receiver points
    lon : number, sequence, or numpy.ndarray
        Longitudes of the receiver points


    Returns
    -------
    altitude: numpy.ndarray
        Topographic altitude (km)


    References
    ----------
    [1] Topography for Earth-to-space propagation modelling:
    https://www.itu.int/rec/R-REC-P.1511/en

    """
    type_output = type(lat)
    lat = prepare_input_array(lat)
    lon = prepare_input_array(lon)
    lon = np.mod(lon, 360)
    val = __model.topographic_altitude(lat, lon)
    val = np.maximum(val, 1e-7)
    return prepare_output_array(val, type_output) * u.km



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


import numpy as np
from astropy import units as u

from itur import utils
from itur.models.itu1144 import bilinear_2D_interpolator
from itur.utils import (prepare_input_array, prepare_output_array,
                        load_data_interpolator)


class __ITU839__():

    """Rain height model for prediction methods.

    Not available versions:
    * P.839-0 (03/92) (Superseded)
    * P.839-1 (83/97) (Superseded)
    Available versions include:
    * P.839-2 (10/99) (Superseded)
    * P.839-3 (02/01) (Superseded)
    * P.839-4 (09/2013) (Current version)
    """
    # This is an abstract class that contains an instance to a version of the
    # ITU-R P.839 recommendation.

    def __init__(self, version=4):
        if version == 4:
            self.instance = _ITU839_4_()
        elif version == 3:
            self.instance = _ITU839_3_()
        elif version == 2:
            self.instance = _ITU839_2_()
#        elif version == 1:
#            self.instance = _ITU839_1()
#        elif version == 0:
#            self.instance = _ITU839_0()
        else:
            raise ValueError(
                'Version {0} is not implemented for the ITU-R P.839 model.'
                .format(version))

        self._zero_isoterm_data = {}

    @property
    def __version__(self):
        return self.instance.__version__

    def rain_height(self, lat, lon):
        # Abstract method to compute the rain height
        return self.instance.rain_height(lat, lon)

    def isoterm_0(self, lat, lon):
        # Abstract method to compute the zero isoterm height
        return self.instance.isoterm_0(lat, lon)


class _ITU839_4_():

    def __init__(self):
        self.__version__ = 4
        self.year = 2013
        self.month = 9
        self.link = 'https://www.itu.int/rec/R-REC-P.839/' +\
                    'recommendation.asp?lang=en&parent=R-REC-P.839-4-201309-I'

        self._zero_isoterm_data = {}

    def isoterm_0(self, lat, lon):
        if not self._zero_isoterm_data:
            self._zero_isoterm_data = load_data_interpolator(
                '839/v4_esalat.npz', '839/v4_esalon.npz',
                '839/v4_esa0height.npz', bilinear_2D_interpolator,
                flip_ud=False)

        return self._zero_isoterm_data(
            np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    def rain_height(self, lat_d, lon_d):
        """The rain height is computed as

        ..math:
            h_r = h_0 + 0.36 (km)
        """
        return self.isoterm_0(lat_d, lon_d) + 0.36


class _ITU839_3_():

    def __init__(self):
        self.__version__ = 3
        self.year = 2001
        self.month = 2
        self.link = 'https://www.itu.int/rec/R-REC-P.839-3-200102-S/en'

        self._zero_isoterm_data = {}

    def isoterm_0(self, lat, lon):
        if not self._zero_isoterm_data:
            self._zero_isoterm_data =  load_data_interpolator(
                '839/v3_esalat.npz', '839/v3_esalon.npz',
                '839/v3_esa0height.npz', bilinear_2D_interpolator,
                flip_ud=False)

        return self._zero_isoterm_data(
            np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    def rain_height(self, lat_d, lon_d):
        """
        The rain height is computed as

        ..math:
            h_r = h_0 + 0.36 (km)
        """
        return self.isoterm_0(lat_d, lon_d) + 0.36


class _ITU839_2_():

    def __init__(self):
        self.__version__ = 2
        self.year = 1999
        self.month = 10
        self.link = 'https://www.itu.int/rec/R-REC-P.839-2-199910-S/en'

    @staticmethod
    def isoterm_0(lat_d, lon_d):
        """
        The 0C mean isotherm height can be approximated as

        """
        # TODO: Complete this with the equation in ITU-R P.839-2
        h0 = np.where(
            lat_d > 23, 5 - 0.075 * (lat_d - 23),
            np.where(
                np.logical_and(0 < lat_d, lat_d < 23),
                5, np.where(
                    np.logical_and(-21 < lat_d, lat_d < 0),
                    5, np.where(
                        np.logical_and(-71 < lat_d, lat_d < -21),
                        5 + 0.1 * (lat_d + 21),
                        0))))
        return h0

    def rain_height(self, lat_d, lon_d):
        """
        For areas of the world where no specific information is available,
        the mean rain height, may be approximated by the mean 0C isotherm
        height, and for for North America and for Europe west of 60° E
        longitude the mean rain height is approximated by

        ..math:
            h_r = 3.2 - 0.075 (\\lambda - 35) \\qquad for \\qquad
            35 \\le \\lambda \\le 70  (km)
        """
        h0 = self.isoterm_0(lat_d, lon_d)
        return np.where(np.logical_and(np.logical_and(35 < lat_d, lat_d < 70),
                                       lon_d < 60),
                        3.2 - 0.075 * (lat_d - 35), h0)


__model = __ITU839__()


def change_version(new_version):
    """
    Change the version of the ITU-R P.839 recommendation currently being used.

    This function changes the model used for the ITU-R P.839 recommendation
    to a different version.

    Parameters
    ----------
    new_version : int
        Number of the version to use.
        Valid values are:
          * 4: Activates recommendation ITU-R P.839-4 (09/2013) (Current version)
          * 3: Activates recommendation ITU-R P.839-3 (02/01) (Superseded)
          * 2: Activates recommendation ITU-R P.839-2 (10/99) (Superseded)


    """
    global __model
    __model = __ITU839__(new_version)


def get_version():
    """
    Obtain the version of the ITU-R P.839 recommendation currently being used.

    Returns
    -------
    version: int
        Version currently being used.
    """
    return __model.__version__


def isoterm_0(lat, lon):
    """
    Estimate the zero degree Celsius isoterm height for propagation prediction.


    Parameters
    ----------
    lat : number, sequence, or numpy.ndarray
        Latitudes of the receiver points
    lon : number, sequence, or numpy.ndarray
        Longitudes of the receiver points


    Returns
    -------
    h0: numpy.ndarray
        Zero degree Celsius isoterm height (km)


    References
    ----------
    [1] Rain height model for prediction methods:
    https://www.itu.int/rec/R-REC-P.839/en

    """
    type_output = type(lat)
    lat = prepare_input_array(lat)
    lon = prepare_input_array(lon)
    lon = np.mod(lon, 360)
    val = __model.isoterm_0(lat, lon)
    return prepare_output_array(val, type_output) * u.km


def rain_height(lat, lon):
    """
    Estimate the annual mean rain height for propagation prediction.

    The mean annual rain height above mean sea level, :math:`h_R`,
    may be obtained from the 0° C isotherm as:

    .. math::

      h_R = h_0 + 0.36 \\qquad \\text{km}


    Parameters
    ----------
    lat : number, sequence, or numpy.ndarray
        Latitudes of the receiver points
    lon : number, sequence, or numpy.ndarray
        Longitudes of the receiver points


    Returns
    -------
    hR: numpy.ndarray
        Annual mean rain height (km)


    References
    ----------
    [1] Rain height model for prediction methods:
    https://www.itu.int/rec/R-REC-P.839/en
    """
    type_output = type(lat)
    lat = prepare_input_array(lat)
    lon = prepare_input_array(lon)
    lon = np.mod(lon, 360)
    val = __model.rain_height(lat, lon)
    return prepare_output_array(val, type_output) * u.km



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import scipy.stats as stats
import scipy.special
import scipy.integrate
from astropy import units as u

from itur import utils
from itur.models.itu453 import water_vapour_pressure,\
    wet_term_radio_refractivity, map_wet_term_radio_refractivity
from itur.models.itu837 import rainfall_rate, rainfall_probability
from itur.models.itu838 import rain_specific_attenuation
from itur.models.itu839 import rain_height
from itur.models.itu1511 import topographic_altitude
from itur.utils import prepare_input_array, prepare_output_array,\
    prepare_quantity, compute_distance_earth_to_earth

import warnings


def __CDF_bivariate_normal__(alpha_x, alpha_y, rho):
    # This function calculates the complementary bivariate normal
    # distribution with limits alpha_x, alpha_y and correlation factor rho
    def CDF_bivariate_normal_fcn(x, y, rho):
        return np.exp(- (x**2 - 2 * rho * x * y + y**2) /
                      (2. * (1 - rho**2)))

    def CDF_bivariate_normal_int(alpha, y, rho):
        return scipy.integrate.quad(
            CDF_bivariate_normal_fcn, alpha, np.inf, args=(y, rho))[0]

    return 1 / (2 * np.pi * np.sqrt(1 - rho**2)) * scipy.integrate.quad(
        lambda y: CDF_bivariate_normal_int(alpha_x, y, rho),
        alpha_y,
        np.inf)[0]


class _ITU618():
    """
    Propagation data and prediction methods required for the design of
    Earth-space telecommunication systems.

    Available versions include:
       * P.618-13 (12/17) (Current version)
       * P.618-12 (07/15) (Superseded)

    Versions that need to be implemented
       * P.618-11
       * P.618-10
       * P.618-09
       * P.618-08
       * P.618-07
       * P.618-06
       * P.618-05
       * P.618-04
       * P.618-03
       * P.618-02
       * P.618-01

    Recommendation ITU-R P.618 provides methods to estimate the propagation
    loss on an Earth-space path, relative to the free-space loss. This value
    is the sum of different contributions as follows:
      * attenuation by atmospheric gases;
      * attenuation by rain, other precipitation and clouds;
      * focusing and defocusing;
      * decrease in antenna gain due to wave-front incoherence;
      * scintillation and multipath effects;
      * attenuation by sand and dust storms.

    Each of these contributions has its own characteristics as a function of
    frequency, geographic location and elevation angle. As a rule, at elevation
    angles above 10°, only gaseous attenuation, rain and cloud attenuation and
    possibly scintillation will be significant, depending on propagation
    conditions.
    """
    # This is an abstract class that contains an instance to a version of the
    # ITU-R P.618 recommendation.

    def __init__(self, version=13):
        if version == 13:
            self.instance = _ITU618_13()
        elif version == 12:
            self.instance = _ITU618_12()
#        elif version == 11:
#            self.instance = _ITU618_11()
#        elif version == 10:
#            self.instance = _ITU618_10()
#        elif version == 9:
#            self.instance = _ITU618_9()
#        elif version == 8:
#            self.instance = _ITU618_8()
#        elif version == 7:
#            self.instance = _ITU618_7()
#        elif version == 6:
#            self.instance = _ITU618_6()
#        elif version == 5:
#            self.instance = _ITU618_5()
#        elif version == 4:
#            self.instance = _ITU618_4()
#        elif version == 3:
#            self.instance = _ITU618_3()
#        elif version == 2:
#            self.instance = _ITU618_2()
#        elif version == 1:
#            self.instance = _ITU618_1()
        else:
            raise ValueError(('Version {0} is not implemented'
                              ' for the ITU-R P.618 model.').format(version))

    @property
    def __version__(self):
        return self.instance.__version__

    def rain_attenuation(self, lat, lon, f, el, hs=None, p=0.01, R001=None,
                         tau=45, Ls=None):
        fcn = np.vectorize(self.instance.rain_attenuation,
                           excluded=[0, 1, 3, 4, 6], otypes=[np.ndarray])
        return np.array(fcn(lat, lon, f, el, hs, p, R001, tau, Ls).tolist())

    def rain_attenuation_probability(self, lat, lon, el, hs, Ls, P0=None):
        fcn = np.vectorize(self.instance.rain_attenuation_probability,
                           excluded=[0, 1, 2], otypes=[np.ndarray])
        return np.array(fcn(lat, lon, el, hs, Ls, P0).tolist())

    def rain_cross_polarization_discrimination(self, Ap, f, el, p, tau):
        fcn = np.vectorize(
            self.instance.rain_cross_polarization_discrimination)
        return fcn(Ap, f, el, p, tau)

    def scintillation_attenuation(self, lat, lon, f, el, p, D, eta,
                                  T, H, P, hL):
        fcn = np.vectorize(self.instance.scintillation_attenuation,
                           excluded=[0, 1, 3, 7, 8, 9], otypes=[np.ndarray])
        return np.array(fcn(lat, lon, f, el, p, D, eta, T, H, P, hL).tolist())

    def scintillation_attenuation_sigma(self, lat, lon, f, el, p, D, eta,
                                        T, H, P, hL):
        fcn = np.vectorize(self.instance.scintillation_attenuation_sigma,
                           excluded=[0, 1, 3, 7, 8, 9], otypes=[np.ndarray])
        return np.array(fcn(lat, lon, f, el, p, D, eta, T, H, P, hL).tolist())

    def fit_rain_attenuation_to_lognormal(self, lat, lon, f, el, hs, P_k, tau):
        fcn = np.vectorize(self.instance.fit_rain_attenuation_to_lognormal)
        return fcn(lat, lon, f, el, hs, P_k, tau)

    def site_diversity_rain_outage_probability(self, lat1, lon1, a1, el1,
                                               lat2, lon2, a2, el2, f, tau=45,
                                               hs1=None, hs2=None):
        fcn = np.vectorize(
                self.instance.site_diversity_rain_outage_probability)
        return np.array(fcn(lat1, lon1, a1, el1,
                            lat2, lon2, a2, el2,
                            f, tau, hs1, hs2).tolist())


class _ITU618_13():

    def __init__(self):
        self.__version__ = 13

    @classmethod
    def rain_attenuation(self, lat, lon, f, el, hs=None, p=0.01, R001=None,
                         tau=45, Ls=None):
        if np.logical_or(p < 0.001, p > 5).any():
            warnings.warn(
                RuntimeWarning('The method to compute the rain attenuation in '
                               'recommendation ITU-P 618-12 is only valid for '
                               'unavailability values between 0.001 and 5'))

        Re = 8500   # Efective radius of the Earth (8500 km)

        if hs is None:
            hs = topographic_altitude(lat, lon).to(u.km).value

        # Step 1: Compute the rain height (hr) based on ITU - R P.839
        hr = rain_height(lat, lon).value

        # Step 2: Compute the slant path length
        if Ls is None:
            Ls = np.where(
                el >= 5, (hr - hs) / (np.sin(np.deg2rad(el))),         # Eq. 1
                2 * (hr - hs) / (((np.sin(np.deg2rad(el)))**2 +
                2 * (hr - hs) / Re)**0.5 + (np.sin(np.deg2rad(el)))))  # Eq. 2

        # Step 3: Calculate the horizontal projection, LG, of the
        # slant-path length
        Lg = np.abs(Ls * np.cos(np.deg2rad(el)))

        # Obtain the raingall rate, exceeded for 0.01% of an average year,
        # if not provided, as described in ITU-R P.837.
        if R001 is None:
            R001 = rainfall_rate(lat, lon, 0.01).to(u.mm / u.hr).value

        # Step 5: Obtain the specific attenuation gammar using the frequency
        # dependent coefficients as given in ITU-R P.838
        # https://www.itu.int/dms_pubrec/itu-r/rec/p/R-REC-P.838-3-200503-I!!PDF-E.pdf
        gammar = rain_specific_attenuation(
            R001, f, el, tau).to(
            u.dB / u.km).value

        # Step 6: Calculate the horizontal reduction factor, r0.01,
        # for 0.01% of the time:
        r001 = 1. / (1 + 0.78 * np.sqrt(Lg * gammar / f) -
                     0.38 * (1 - np.exp(-2 * Lg)))

        # Step 7: Calculate the vertical adjustment factor, v0.01,
        # for 0.01% of the time:
        eta = np.rad2deg(np.arctan2(hr - hs, Lg * r001))

        Lr = np.where(eta > el, Lg * r001 / np.cos(np.deg2rad(el)),
                      (hr - hs) / np.sin(np.deg2rad(el)))

        xi = np.where(np.abs(lat) < 36, 36 - np.abs(lat), 0)

        v001 = 1. / (1 + np.sqrt(np.sin(np.deg2rad(el))) *
                     (31 * (1 - np.exp(-(el / (1 + xi)))) *
                      np.sqrt(Lr * gammar) / f**2 - 0.45))

        # Step 8: calculate the effective path length:
        Le = Lr * v001   # (km)

        # Step 9: The predicted attenuation exceeded for 0.01% of an average
        # year
        A001 = gammar * Le   # (dB)

        # Step 10: The estimated attenuation to be exceeded for other
        # percentages of an average year
        if p >= 1:
            beta = np.zeros_like(A001)
        else:
            beta = np.where(np.abs(lat) > 36,
                            np.zeros_like(A001),
                            np.where((np.abs(lat) < 36) & (el > 25),
                                     -0.005 * (np.abs(lat) - 36),
                                     -0.005 * (np.abs(lat) - 36) + 1.8 -
                                     4.25 * np.sin(np.deg2rad(el))))

        A = A001 * (p / 0.01)**(
            -(0.655 + 0.033 * np.log(p) - 0.045 * np.log(A001) -
              beta * (1 - p) * np.sin(np.deg2rad(el))))

        return A

    @classmethod
    def rain_attenuation_probability(self, lat, lon, el, hs=None,
                                     Ls=None, P0=None):

        Re = 8500
        if hs is None:
            hs = topographic_altitude(lat, lon).to(u.km).value

        # Step 1: Estimate the probability of rain, at the earth station either
        # from Recommendation ITU-R P.837 or from local measured rainfall
        # rate data
        if P0 is None:
            P0 = rainfall_probability(lat, lon).\
                to(u.dimensionless_unscaled).value

        # Step 2: Calculate the parameter alpha using the inverse of the
        # Q-function alpha = Q^{-1}(P0)  -> Q(alpha) = P0
        alpha = stats.norm.ppf(1 - P0)

        # Step 3: Calculate the spatial correlation function, rho:
        hr = rain_height(lat, lon).value

        if Ls is None:
            Ls = np.where(
                el >= 5, (hr - hs) / (np.sin(np.deg2rad(el))),         # Eq. 1
                2 * (hr - hs) / (((np.sin(np.deg2rad(el)))**2 +
                2 * (hr - hs) / Re)**0.5 + (np.sin(np.deg2rad(el)))))  # Eq. 2

        d = Ls * np.cos(np.deg2rad(el))
        rho = 0.59 * np.exp(-abs(d) / 31) + 0.41 * np.exp(-abs(d) / 800)

        # Step 4: Calculate the complementary bivariate normal distribution
        biva_fcn = np.vectorize(__CDF_bivariate_normal__)
        c_B = biva_fcn(alpha, alpha, rho)

        # Step 5: Calculate the probability of rain attenuation on the slant
        # path:
        P = 1 - (1 - P0) * ((c_B - P0**2) / (P0 * (1 - P0)))**P0
        return P

    @classmethod
    def fit_rain_attenuation_to_lognormal(self, lat, lon, f, el, hs, P_k, tau):
        # Performs the log-normal fit of rain attenuation vs. probability of
        # occurrence for a particular path

        # Step 1: Construct the set of pairs [Pi, Ai] where Pi (% of time) is
        # the probability the attenuation Ai (dB) is exceeded where Pi < P_K
        p_i = np.array([0.01, 0.02, 0.03, 0.05,
                        0.1, 0.2, 0.3, 0.5, 1, 2, 3, 5, 10])
        Pi = np.array([p for p in p_i if p < P_k], dtype=np.float)
        Ai = np.array([0 for p in p_i if p < P_k], dtype=np.float)

        for i, p in enumerate(Pi):
            Ai[i] = self.rain_attenuation(lat, lon, f, el, hs, p, tau=tau)

        # Step 2: Transform the set of pairs [Pi, Ai] to [Q^{-1}(Pi/P_k),
        # ln(Ai)]
        Q = stats.norm.ppf(1 - (Pi / P_k))
        lnA = np.log(Ai)

        # Step 3: Determine the variables sigma_lna, m_lna by performing a
        # least-squares fit to lnAi = sigma_lna Q^{-1}(Pi/P_k) + m_lna
        m_lna, sigma_lna = np.linalg.lstsq(np.vstack([np.ones(len(Q)), Q]).T,
                                           lnA, rcond=None)[0]

        return sigma_lna, m_lna

    @classmethod
    def site_diversity_rain_outage_probability(self, lat1, lon1, a1, lat2,
                                               lon2, a2, f, el1, el2, tau=45,
                                               hs1=None, hs2=None):
        # The diversity prediction method assumes a log-normal distribution of
        # rain intensity and rain attenuation. This method predicts
        # Pr(A1 > a1, A2 > a2), the joint probability (%) that the attenuation
        # on the path to the first site is greater than a1 and the attenuation
        # on the path to the second site is greater than a2.
        d = compute_distance_earth_to_earth(lat1, lon1, lat2, lon2)
        rho_r = 0.7 * np.exp(-d / 60) + 0.3 * np.exp(-(d / 700)**2)

        P_1 = rainfall_probability(lat1, lon1).\
            to(u.dimensionless_unscaled).value

        P_2 = rainfall_probability(lat2, lon2).\
            to(u.dimensionless_unscaled).value

        R_1 = stats.norm.ppf(1 - P_1)
        R_2 = stats.norm.ppf(1 - P_2)
        biva_fcn = np.vectorize(__CDF_bivariate_normal__)
        P_r = biva_fcn(R_1, R_2, rho_r)

        sigma_lna1, m_lna1 = self.fit_rain_attenuation_to_lognormal(
            lat1, lon1, f, el1, hs1, P_1 * 100, tau)

        sigma_lna2, m_lna2 = self.fit_rain_attenuation_to_lognormal(
            lat2, lon2, f, el2, hs1, P_2 * 100, tau)

        rho_a = 0.94 * np.exp(-d / 30) + 0.06 * np.exp(-(d / 500)**2)
        lim_1 = (np.log(a1) - m_lna1) / sigma_lna1
        lim_2 = (np.log(a2) - m_lna2) / sigma_lna2

        P_a = biva_fcn(lim_1, lim_2, rho_a)

        return 100 * P_r * P_a

    @classmethod
    def rain_cross_polarization_discrimination(self, Ap, f, el, p, tau):
        # Frequency reuse by means of orthogonal polarizations is often used to
        # increase the capacity of space telecommunication systems. This
        # technique is restricted, however, by depolarization on atmospheric
        # propagation paths. Various depolarization mechanisms, especially
        # hydrometeor effects, are important in the troposphere

        # The method described below to calculate cross-polarization
        # discrimination (XPD) statistics from rain attenuation statistics for
        # the same path is valid for 6 < f < 55 GHz and el < 60°.
        if f < 4 or f > 55:
            warnings.warn(
                RuntimeWarning(
                    'The method to compute the cross '
                    'polarization discrimination in recommendation '
                    'ITU-P 618-12 is only valid for frequency values between'
                    ' 4 and 55 GHz'))

        if el > 60:
            warnings.warn(
                RuntimeWarning(
                    'The method to compute thecross '
                    'polarization discrimination in recommendation ITU-P '
                    '618-12 is only valid for elevation angle values below '
                    '60 degrees'))

        # In case that the frequency is comprised between 4 and 6 GHz, scaling
        # is necessary
        scale_to_orig_f = False
        if 4 <= f < 6:
            f_orig = f
            f = 6
            scale_to_orig_f = True

        # Step 1: Calculate the frequency-dependent term:
        if 6 <= f < 9:
            C_f = 60 * np.log10(f) - 28.3
        elif 9 <= f < 36:
            C_f = 26 * np.log10(f) + 4.1
        elif 36 <= f <= 55:
            C_f = 35.9 * np.log10(f) - 11.3

        # Step 2: Calculate the rain attenuation dependent term:
        if 6 <= f < 9:
            V = 30.8 * f**-0.21
        elif 9 <= f < 20:
            V = 12.8 * f**0.19
        elif 20 <= f < 40:
            V = 22.6
        elif 40 <= f <= 55:
            V = 13.0 * f**0.15

        C_a = V * np.log10(Ap)

        # Step 3: Calculate the polarization improvement factor:
        C_tau = -10 * np.log10(1 - 0.484 * (1 + np.cos(np.deg2rad(4 * tau))))

        # Step 4: Calculate the elevation angle-dependent term:
        C_theta = -40 * np.log10(np.cos(np.deg2rad(el)))

        # Step 5: Calculate the canting angle dependent term:
        if p <= 0.001:
            C_sigma = 0.0053 * 15**2
        elif p <= 0.01:
            C_sigma = 0.0053 * 10**2
        elif p <= 0.1:
            C_sigma = 0.0053 * 5**2
        else:
            C_sigma = 0

        # Step 6: Calculate rain XPD not exceeded for p% of the time:
        XPD_rain = C_f - C_a + C_tau + C_theta + C_sigma

        # Step 7: Calculate the ice crystal dependent term:
        C_ice = XPD_rain * (0.3 + 0.1 * np.log10(p)) / 2

        # Step 8: Calculate the XPD not exceeded for p% of the time,
        # including the effects of ice:
        XPD_p = XPD_rain - C_ice

        if scale_to_orig_f:
            # Long-term XPD statistics obtained at one frequency and
            # polarization tilt angle can be scaled to another frequency and
            # polarization tilt angle using the semi-empirical formula:
            XPD_p = XPD_p - 20 * np.log10(
              f_orig * np.sqrt(1 - 0.484 * (1 - np.cos(np.deg2rad(4 * tau)))) /
              (f * np.sqrt(1 - 0.484 * (1 - np.cos(np.deg2rad(4 * tau))))))
        return XPD_p

    @classmethod

    def scintillation_attenuation_sigma(self, lat, lon, f, el, p, D, eta=0.5,
                                        T=None, H=None, P=None, hL=1000):
        # Step 1: For the value of t, calculate the saturation water vapour
        # pressure, es, (hPa), as specified in Recommendation ITU-R P.453.
        if T is not None and H is not None and P is not None:
            e = water_vapour_pressure(T, P, H).value

            # Step 2: Compute the wet term of the radio refractivity, Nwet,
            # corresponding to es, t and H as given in Recommendation ITU-R
            # P.453.
            N_wet = wet_term_radio_refractivity(e, T).value
        else:
            N_wet = map_wet_term_radio_refractivity(lat, lon, 50).value

        # Step 3: Calculate the standard deviation of the reference signal
        # amplitude:
        sigma_ref = 3.6e-3 + 1e-4 * N_wet  # Eq. 43   [dB]

        # Step 4: Calculate the effective path length L:
        L = 2 * hL / (np.sqrt(np.sin(np.deg2rad(el))**2 + 2.35e-4) +
                      np.sin(np.deg2rad(el)))  # Eq. 44   [m]

        # Step 5: Estimate the effective antenna diameter, Deff
        D_eff = np.sqrt(eta) * D  # Eq. 45   [m]

        # Step 6: Step 6: Calculate the antenna averaging factor
        x = 1.22 * D_eff**2 * f / L
        g = np.where(x >= 7.0, 0,
                     np.sqrt(3.86 * (x**2 + 1)**(11. / 12) *
                             np.sin(11. / 6 * np.arctan2(1, x)) -
                             7.08 * x**(5. / 6)))  # Eq. 46    [-]

        # Step 7: Calculate the standard deviation of the signal for the
        # applicable period and propagation path:
        sigma = sigma_ref * f**(7. / 12) * g / np.sin(np.deg2rad(el))**1.2
        return sigma

    @classmethod
    def scintillation_attenuation(self, lat, lon, f, el, p, D, eta=0.5, T=None,
                                  H=None, P=None, hL=1000):
        # Step 1 - 7: Calculate the standard deviation of the signal for the
        # applicable period and propagation path:
        sigma = self.scintillation_attenuation_sigma(lat, lon, f, el, p,
                                                     D, eta, T, H, P, hL)
        # Step 8: Calculate the time percentage factor, a(p), for the time
        # percentage, p, in the range between 0.01% < p < 50%:
        a = -0.061 * np.log10(p)**3 + 0.072 * \
            np.log10(p)**2 - 1.71 * np.log10(p) + 3

        # Step 9: Calculate the fade depth, A(p), exceeded for p% of the time:
        A_s = a * sigma  # Eq. 49   [dB]

        return A_s


class _ITU618_12():

    def __init__(self):
        self.__version__ = 12

    @classmethod
    def rain_attenuation(self, lat, lon, f, el, hs=None, p=0.01, R001=None,
                         tau=45, Ls=None):

        if p < 0.001 or p > 5:
            warnings.warn(
                RuntimeWarning('The method to compute the rain attenuation in '
                               'recommendation ITU-P 618-12 is only valid for '
                               'unavailability values between 0.001 and 5'))

        Re = 8500   # Efective radius of the Earth (8500 km)

        if hs is None:
            hs = topographic_altitude(lat, lon).to(u.km).value

        # Step 1: Compute the rain height (hr) based on ITU - R P.839
        hr = rain_height(lat, lon).value

        # Step 2: Compute the slant path length
        if Ls is None:
            Ls = np.where(
                el >= 5, (hr - hs) / (np.sin(np.deg2rad(el))),  # Eq. 1
                2 * (hr - hs) / (((np.sin(np.deg2rad(el)))**2 +
                                  2 * (hr - hs) / Re)**0.5 +
                                 (np.sin(np.deg2rad(el)))))  # Eq. 2

        # Step 3: Calculate the horizontal projection, LG, of the
        # slant-path length
        Lg = np.abs(Ls * np.cos(np.deg2rad(el)))

        # Obtain the raingall rate, exceeded for 0.01% of an average year,
        # if not provided, as described in ITU-R P.837.
        if R001 is None:
            R001 = rainfall_rate(lat, lon, 0.01).to(u.mm / u.hr).value

        # Step 5: Obtain the specific attenuation gammar using the frequency
        # dependent coefficients as given in ITU-R P.838
        gammar = rain_specific_attenuation(
            R001, f, el, tau).to(
            u.dB / u.km).value

        # Step 6: Calculate the horizontal reduction factor, r0.01,
        # for 0.01% of the time:
        r001 = 1. / (1 + 0.78 * np.sqrt(Lg * gammar / f) -
                     0.38 * (1 - np.exp(-2 * Lg)))

        # Step 7: Calculate the vertical adjustment factor, v0.01,
        # for 0.01% of the time:
        eta = np.rad2deg(np.arctan2(hr - hs, Lg * r001))

        Lr = np.where(eta > el, Lg * r001 / np.cos(np.deg2rad(el)),
                      (hr - hs) / np.sin(np.deg2rad(el)))

        xi = np.where(np.abs(lat) < 36, 36 - np.abs(lat), 0)

        v001 = 1. / (1 + np.sqrt(np.sin(np.deg2rad(el))) *
                     (31 * (1 - np.exp(-(el / (1 + xi)))) *
                      np.sqrt(Lr * gammar) / f**2 - 0.45))

        # Step 8: calculate the effective path length:
        Le = Lr * v001   # (km)

        # Step 9: The predicted attenuation exceeded for 0.01% of an average
        # year
        A001 = gammar * Le   # (dB)

        # Step 10: The estimated attenuation to be exceeded for other
        # percentages of an average year
        if p >= 1:
            beta = np.zeros_like(A001)
        else:
            beta = np.where(np.abs(lat) > 36,
                            np.zeros_like(A001),
                            np.where((np.abs(lat) < 36) & (el > 25),
                                     -0.005 * (np.abs(lat) - 36),
                                     -0.005 * (np.abs(lat) - 36) + 1.8 -
                                     4.25 * np.sin(np.deg2rad(el))))

        A = A001 * (p / 0.01)**(-(0.655 + 0.033 * np.log(p) -
                                  0.045 * np.log(A001) -
                                  beta * (1 - p) * np.sin(np.deg2rad(el))))

        return A

    @classmethod
    def rain_attenuation_probability(self, *args, **kwargs):
        return _ITU618_13.rain_attenuation_probability(*args, **kwargs)

    @classmethod
    def fit_rain_attenuation_to_lognormal(self, *args, **kwargs):
        return _ITU618_13.fit_rain_attenuation_to_lognormal(*args, **kwargs)

    @classmethod
    def site_diversity_rain_outage_probability(self, *args, **kwargs):
        return _ITU618_13.site_diversity_rain_outage_probability(*args,
                                                                 **kwargs)

    @classmethod
    def rain_cross_polarization_discrimination(self, *args, **kwargs):
        return _ITU618_13.rain_cross_polarization_discrimination(*args,
                                                                 **kwargs)

    @classmethod
    def scintillation_attenuation(self, *args, **kwargs):
        return _ITU618_13.scintillation_attenuation(*args, **kwargs)


__model = _ITU618()


def change_version(new_version):
    """
    Change the version of the ITU-R P.618 recommendation currently being used.

    This function changes the model used for the ITU-R P.618 recommendation
    to a different version.

    Parameters
    ----------
    new_version : int
        Number of the version to use.
        Valid values are:
          *  13: Activates recommendation ITU-R P.618-13 (12/17) (Current version)
          *  12: Activates recommendation ITU-R P.618-12 (07/15) (Superseded)
    """
    global __model
    __model = _ITU618(new_version)


def get_version():
    """ The version of the current model for the ITU-R P.618 recommendation.

    Obtain the version of the ITU-R P.618 recommendation currently being used.

    Returns
    -------
    version: int
       The version of the ITU-R P.618 recommendation being used.
    """
    return __model.__version__


def rain_attenuation(lat, lon, f, el, hs=None, p=0.01, R001=None,
                     tau=45, Ls=None):
    """
    Calculation of long-term rain attenuation statistics from point rainfall
    rate.
    The following procedure provides estimates of the long-term statistics of
    the slant-path rain attenuation at a given location for frequencies up
    to 55 GHz.


    Parameters
    ----------
    lat : number, sequence, or numpy.ndarray
        Latitudes of the receiver points
    lon : number, sequence, or numpy.ndarray
        Longitudes of the receiver points
    f : number
        Frequency (GHz)
    el : sequence, or number
        Elevation angle (degrees)
    hs : number, sequence, or numpy.ndarray, optional
        Heigh above mean sea level of the earth station (km). If local data for
        the earth station height above mean sea level is not available, an
        estimate is obtained from the maps of topographic altitude
        given in Recommendation ITU-R P.1511.
    p : number, optional
        Percentage of the time the rain attenuation value is exceeded.
    R001: number, optional
        Point rainfall rate for the location for 0.01% of an average year
        (mm/h).
        If not provided, an estimate is obtained from Recommendation
        Recommendation ITU-R P.837. Some useful values:
            * 0.25 mm/h : Drizzle
            *  2.5 mm/h : Light rain
            * 12.5 mm/h : Medium rain
            * 25.0 mm/h : Heavy rain
            * 50.0 mm/h : Downpour
            * 100  mm/h : Tropical
            * 150  mm/h : Monsoon
    tau : number, optional
        Polarization tilt angle relative to the horizontal (degrees)
        (tau = 45 deg for circular polarization). Default value is 45
    Ls :number, optional
        Slant path length (km). If not provided, it will be computed using the
        rain height and the elevation angle. The ITU model does not require
        this parameter as an input.


    Returns
    -------
    attenuation: Quantity
        Attenuation due to rain (dB)

    References
    --------
    [1] Propagation data and prediction methods required for the design of
    Earth-space telecommunication systems:
    https://www.itu.int/dms_pubrec/itu-r/rec/p/R-REC-P.618-12-201507-I!!PDF-E.pdf
    """
    type_output = type(lat)

    lat = prepare_input_array(lat)
    lon = prepare_input_array(lon)

    lon = np.mod(lon, 360)
    f = prepare_quantity(f, u.GHz, 'Frequency')
    el = prepare_quantity(prepare_input_array(el), u.deg, 'Elevation angle')
    hs = prepare_quantity(
        hs, u.km, 'Heigh above mean sea level of the earth station')
    R001 = prepare_quantity(R001, u.mm / u.hr, 'Point rainfall rate')
    tau = prepare_quantity(tau, u.one, 'Polarization tilt angle')
    Ls = prepare_quantity(Ls, u.km, 'Slant path length')

    val = __model.rain_attenuation(lat, lon, f, el, hs=hs, p=p,
                                   R001=R001, tau=tau, Ls=Ls)
    return prepare_output_array(val, type_output) * u.dB


def rain_attenuation_probability(lat, lon, el, hs=None, Ls=None, P0=None):
    """
    The following procedure computes the probability of non-zero rain
    attenuation on a given slant path Pr(Ar > 0).


    Parameters
    ----------
    lat : number, sequence, or numpy.ndarray
        Latitudes of the receiver points
    lon : number, sequence, or numpy.ndarray
        Longitudes of the receiver points
    el : sequence, or number
        Elevation angle (degrees)
    hs : number, sequence, or numpy.ndarray, optional
        Heigh above mean sea level of the earth station (km). If local data for
        the earth station height above mean sea level is not available, an
        estimate is obtained from the maps of topographic altitude
        given in Recommendation ITU-R P.1511.
    Ls : number, sequence, or numpy.ndarray, optional
        Slant path length from the earth station to the rain height (km). If
        data about the rain height is not available, this value is estimated
        automatically using Recommendation ITU-R P.838
    P0 : number, sequence, or numpy.ndarray, optional
        Probability of rain at the earth station, (0 ≤ P0 ≤ 1)



    Returns
    -------
    p: Quantity
         Probability of rain attenuation on the slant path (%)


    References
    ----------
    [1] Propagation data and prediction methods required for the design of
    Earth-space telecommunication systems:
    https://www.itu.int/dms_pubrec/itu-r/rec/p/R-REC-P.618-12-201507-I!!PDF-E.pdf
    """
    type_output = type(lat)

    lat = prepare_input_array(lat)
    lon = prepare_input_array(lon)

    lon = np.mod(lon, 360)
    el = prepare_quantity(prepare_input_array(el), u.deg, 'Elevation angle')
    hs = prepare_quantity(
        hs, u.km, 'Heigh above mean sea level of the earth station')
    Ls = prepare_quantity(
        Ls, u.km, 'Heigh above mean sea level of the earth station')
    P0 = prepare_quantity(P0, u.pct, 'Point rainfall rate')

    val = __model.rain_attenuation_probability(lat, lon, el, hs, Ls, P0)

    return prepare_output_array(val, type_output) * 100 * u.pct


def site_diversity_rain_outage_probability(lat1, lon1, a1, el1, lat2,
                                           lon2, a2, el2, f, tau=45, hs1=None,
                                           hs2=None):
    """
    Calculate the link outage probability in a diversity based scenario (two
    ground stations) due to rain attenuation. This method is valid for
    frequencies below 20 GHz, as at higher frequencies other impairments might
    affect affect site diversity performance.

    This method predicts Pr(A1 > a1, A2 > a2), the joint probability (%) that
    the attenuation on the path to the first site is greater than a1 and the
    attenuation on the path to the second site is greater than a2.


    Parameters
    ----------
    lat1 : number or Quantity
        Latitude of the first ground station (deg)
    lon1 : number or Quantity
        Longitude of the first ground station (deg)
    a1 : number or Quantity
        Maximum admissible attenuation of the first ground station (dB)
    el1 : number or Quantity
        Elevation angle to the first ground station (deg)
    lat2 : number or Quantity
        Latitude of the second ground station (deg)
    lon2 : number or Quantity
        Longitude of the second ground station (deg)
    a2 : number or Quantity
        Maximum admissible attenuation of the second ground station (dB)
    el2 : number or Quantity
        Elevation angle to the second ground station (deg)
    f : number or Quantity
        Frequency (GHz)
    tau : number, optional
        Polarization tilt angle relative to the horizontal (degrees)
        (tau = 45 deg for circular polarization). Default value is 45
    hs1 : number or Quantity, optional
        Altitude over the sea level of the first ground station (km). If not
        provided, uses Recommendation ITU-R P.1511 to compute the toporgraphic
        altitude
    hs2 : number or Quantity, optional
        Altitude over the sea level of the first ground station (km). If not
        provided, uses Recommendation ITU-R P.1511 to compute the toporgraphic
        altitude


    Returns
    -------
    probability: Quantity
        Joint probability (%) that the attenuation on the path to the first
        site is greater than a1 and the attenuation on the path to the second
        site is greater than a2


    References
    ----------
    [1] Propagation data and prediction methods required for the design of
    Earth-space telecommunication systems:
    https://www.itu.int/dms_pubrec/itu-r/rec/p/R-REC-P.618-12-201507-I!!PDF-E.pdf
    """
    type_output = type(lat1)
    lon1 = np.mod(lon1, 360)
    lat1 = prepare_quantity(lat1, u.deg, 'Latitude in ground station 1')
    lon1 = prepare_quantity(lon1, u.deg, 'Longitude in ground station 1')
    a1 = prepare_quantity(a1, u.dB, 'Attenuation margin in ground station 1')

    lon2 = np.mod(lon2, 360)
    lat2 = prepare_quantity(lat2, u.deg, 'Latitude in ground station 2')
    lon2 = prepare_quantity(lon2, u.deg, 'Longitude in ground station 2')
    a2 = prepare_quantity(a2, u.dB, 'Attenuation margin in ground station 2')

    f = prepare_quantity(f, u.GHz, 'Frequency')
    tau = prepare_quantity(tau, u.one, 'Polarization tilt angle')
    el1 = prepare_quantity(el1, u.deg, 'Elevation angle in ground station 1')
    el2 = prepare_quantity(el2, u.deg, 'Elevation angle in ground station 2')
    hs1 = prepare_quantity(
        hs1, u.km, 'Altitude over the sea level for ground station 1')
    hs2 = prepare_quantity(
        hs2, u.km, 'Altitude over the sea level for ground station 2')

    val = __model.site_diversity_rain_outage_probability(
        lat1, lon1, a1, lat2, lon2, a2, f, el1, el2, tau=tau, hs1=hs1, hs2=hs2)

    return prepare_output_array(val, type_output) * u.pct


def scintillation_attenuation(lat, lon, f, el, p, D, eta=0.5, T=None,
                              H=None, P=None, hL=1000):
    """
    Calculation of monthly and long-term statistics of amplitude scintillations
    at elevation angles greater than 5° and frequencies up to 20 GHz.


    Parameters
    ----------
    lat : number, sequence, or numpy.ndarray
        Latitudes of the receiver points
    lon : number, sequence, or numpy.ndarray
        Longitudes of the receiver points
    f : number or Quantity
        Frequency (GHz)
    el : sequence, or number
        Elevation angle (degrees)
    p : number
        Percentage of the time the scintillation attenuation value is exceeded.
    D: number
        Physical diameter of the earth-station antenna (m)
    eta: number, optional
        Antenna efficiency. Default value 0.5 (conservative estimate)
    T: number, sequence, or numpy.ndarray, optional
        Average surface ambient temperature (°C) at the site. If None, uses the
        ITU-R P.453 to estimate the wet term of the radio refractivity.
    H: number, sequence, or numpy.ndarray, optional
        Average surface relative humidity (%) at the site. If None, uses the
        ITU-R P.453 to estimate the wet term of the radio refractivity.
    P: number, sequence, or numpy.ndarray, optional
        Average surface pressure (hPa) at the site. If None, uses the
        ITU-R P.453 to estimate the wet term of the radio refractivity.
    hL : number, optional
        Height of the turbulent layer (m). Default value 1000 m


    Returns
    -------
    attenuation: Quantity
        Attenuation due to scintillation (dB)


    References
    ----------
    [1] Propagation data and prediction methods required for the design of
    Earth-space telecommunication systems:
    https://www.itu.int/dms_pubrec/itu-r/rec/p/R-REC-P.618-12-201507-I!!PDF-E.pdf
    """
    type_output = type(lat)

    lat = prepare_input_array(lat)
    lon = prepare_input_array(lon)

    lon = np.mod(lon, 360)
    f = prepare_quantity(f, u.GHz, 'Frequency')
    el = prepare_quantity(prepare_input_array(el), u.deg, 'Elevation angle')
    D = prepare_quantity(D, u.m, 'Antenna diameter')
    eta = prepare_quantity(eta, u.one, 'Antenna efficiency')
    T = prepare_quantity(T, u.deg_C, 'Average surface temperature')
    H = prepare_quantity(H, u.percent, 'Average surface relative humidity')
    P = prepare_quantity(P, u.hPa, 'Average surface pressure')
    hL = prepare_quantity(hL, u.m, 'Height of the turbulent layer')

    val = __model.scintillation_attenuation(
        lat, lon, f, el, p, D, eta=eta, T=T, H=H, P=P, hL=hL)

    return prepare_output_array(val, type_output) * u.dB


def scintillation_attenuation_sigma(lat, lon, f, el, p, D, eta=0.5, T=None,
                                    H=None, P=None, hL=1000):
    """
    Calculation of the standard deviation of the amplitude of the
    scintillations attenuation at elevation angles greater than 5° and
    frequencies up to 20 GHz.


    Parameters
    ----------
    lat : number, sequence, or numpy.ndarray
        Latitudes of the receiver points
    lon : number, sequence, or numpy.ndarray
        Longitudes of the receiver points
    f : number or Quantity
        Frequency (GHz)
    el : sequence, or number
        Elevation angle (degrees)
    p : number
        Percentage of the time the scintillation attenuation value is exceeded.
    D: number
        Physical diameter of the earth-station antenna (m)
    eta: number, optional
        Antenna efficiency. Default value 0.5 (conservative estimate)
    T: number, sequence, or numpy.ndarray, optional
        Average surface ambient temperature (°C) at the site. If None, uses the
        ITU-R P.453 to estimate the wet term of the radio refractivity.
    H: number, sequence, or numpy.ndarray, optional
        Average surface relative humidity (%) at the site. If None, uses the
        ITU-R P.453 to estimate the wet term of the radio refractivity.
    P: number, sequence, or numpy.ndarray, optional
        Average surface pressure (hPa) at the site. If None, uses the
        ITU-R P.453 to estimate the wet term of the radio refractivity.
    hL : number, optional
        Height of the turbulent layer (m). Default value 1000 m


    Returns
    -------
    attenuation: Quantity
        Attenuation due to scintillation (dB)


    References
    ----------
    [1] Propagation data and prediction methods required for the design of
    Earth-space telecommunication systems:
    https://www.itu.int/dms_pubrec/itu-r/rec/p/R-REC-P.618-12-201507-I!!PDF-E.pdf
    """
    type_output = type(lat)

    lat = prepare_input_array(lat)
    lon = prepare_input_array(lon)

    lon = np.mod(lon, 360)
    f = prepare_quantity(f, u.GHz, 'Frequency')
    el = prepare_quantity(prepare_input_array(el), u.deg, 'Elevation angle')
    D = prepare_quantity(D, u.m, 'Antenna diameter')
    eta = prepare_quantity(eta, u.one, 'Antenna efficiency')
    T = prepare_quantity(T, u.deg_C, 'Average surface temperature')
    H = prepare_quantity(H, u.percent, 'Average surface relative humidity')
    P = prepare_quantity(P, u.hPa, 'Average surface pressure')
    hL = prepare_quantity(hL, u.m, 'Height of the turbulent layer')

    val = __model.scintillation_attenuation_sigma(
        lat, lon, f, el, p, D, eta=eta, T=T, H=H, P=P, hL=hL)

    return prepare_output_array(val, type_output) * u.dB


def rain_cross_polarization_discrimination(Ap, f, el, p, tau=45):
    """
    Calculation of the cross-polarization discrimination (XPD) statistics from
    rain attenuation statistics. The following procedure provides estimates of
    the long-term statistics of the cross-polarization discrimination (XPD)
    statistics for frequencies up to 55 GHz and elevation angles lower than 60
    deg.


    Parameters
    ----------
    Ap : number, sequence, or numpy.ndarray
        Rain attenuation (dB) exceeded for the required percentage of time, p,
        for the path in question, commonly called co-polar attenuation (CPA)
    f : number
        Frequency
    el : number, sequence, or numpy.ndarray
        Elevation angle (degrees)
    p : number
        Percentage of the time the XPD is exceeded.
    tau : number, optional
        Polarization tilt angle relative to the horizontal (degrees)
        (tau = 45 deg for circular polarization). Default value is 45


    Returns
    -------
    attenuation: Quantity
        Cross-polarization discrimination (dB)


    References
    ----------
    [1] Propagation data and prediction methods required for the design of
    Earth-space telecommunication systems:
    https://www.itu.int/dms_pubrec/itu-r/rec/p/R-REC-P.618-12-201507-I!!PDF-E.pdf
    """
    type_output = type(Ap)
    Ap = prepare_input_array(Ap)
    f = prepare_quantity(f, u.GHz, 'Frequency')
    el = prepare_quantity(el, u.deg, 'Elevation angle')
    tau = prepare_quantity(tau, u.one, 'Polarization tilt angle')
    val = __model.rain_cross_polarization_discrimination(Ap, f, el, p, tau=tau)
    return prepare_output_array(val, type_output) * u.dB


def fit_rain_attenuation_to_lognormal(lat, lon, f, el, hs, P_k, tau):
    """
    Compute the log-normal fit of rain attenuation vs. probability of
    occurrence.


    Parameters
    ----------
    lat : number, sequence, or numpy.ndarray
        Latitudes of the receiver points
    lon : number, sequence, or numpy.ndarray
        Longitudes of the receiver points
    f : number or Quantity
        Frequency (GHz)
    el : sequence, or number
        Elevation angle (degrees)
    hs : number, sequence, or numpy.ndarray, optional
        Heigh above mean sea level of the earth station (km). If local data for
        the earth station height above mean sea level is not available, an
        estimate is obtained from the maps of topographic altitude
        given in Recommendation ITU-R P.1511.
    P_k : number
        Rain probability
    tau : number, optional
        Polarization tilt angle relative to the horizontal (degrees)
        (tau = 45 deg for circular polarization). Default value is 45

    Returns
    -------
    sigma_lna:
        Standar deviation of the lognormal distribution
    m_lna:
        Mean of the lognormal distribution

    References
    ----------
    [1] Propagation data and prediction methods required for the design of
    Earth-space telecommunication systems:
    https://www.itu.int/dms_pubrec/itu-r/rec/p/R-REC-P.618-12-201507-I!!PDF-E.pdf

    """
    lat = prepare_input_array(lat)
    lon = prepare_input_array(lon)
    lon = np.mod(lon, 360)
    f = prepare_quantity(f, u.GHz, 'Frequency')
    el = prepare_quantity(prepare_input_array(el), u.deg, 'Elevation angle')
    hs = prepare_quantity(
        hs, u.km, 'Heigh above mean sea level of the earth station')
    tau = prepare_quantity(tau, u.one, 'Polarization tilt angle')
    sigma_lna, m_lna = __model.fit_rain_attenuation_to_lognormal(
        lat, lon, f, el, hs, P_k, tau)
    return sigma_lna, m_lna



# -*- coding: utf-8 -*-
"""
Interpolation methods for the geophysical properties used to compute
propagation effects. These methods are based on those in Recommendation
ITU-R P.1144-7.

References
--------
[1] Guide to the application of the propagation methods of Radiocommunication
Study Group 3: https://www.itu.int/rec/R-REC-P.1144/en
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
from scipy.interpolate import griddata, RegularGridInterpolator


def is_regular_grid(lats_o, lons_o):
    """
    Determinere whether the grids in lats_o and lons_o are both regular grids
    or not.

    A grid is regular if the difference (column-wise or row-wise)
    between consecutive values is constant across the grid.


    Parameters
    -----------
    lats_o : numpy.ndarray
        Grid of latitude coordinates
    lons_o : numpy.ndarray
        Grid of longitude coordinates


    Returns
    --------
        is_regular: boolean
    """
    Delta_lons = np.unique(np.diff(lons_o, axis=1))
    Delta_lats = np.unique(np.diff(lats_o, axis=0))

    return (np.allclose(Delta_lons, Delta_lons[0], rtol=1e-5) and
            np.allclose(Delta_lats, Delta_lats[0], rtol=1e-5) and
            (Delta_lons != 0).all() and (Delta_lats != 0).all())

###############################################################################
#                       Nearest Neighbour Interpolation                       #
###############################################################################


def nearest_2D_interpolator(lats_o, lons_o, values):
    """
    Produces a 2D interpolator function using the nearest value interpolation
    method. If the grids are regular grids, uses the
    scipy.interpolate.RegularGridInterpolator,
    otherwise, scipy.intepolate.griddata

    Values can be interpolated from the returned function as follows:

    .. code-block:: python

      f = nearest_2D_interpolator(lat_origin, lon_origin, values_origin)
      interp_values = f(lat_interp, lon_interp)


    Parameters
    -----------
    lats_o: numpy.ndarray
        Latitude coordinates of the values usde by the interpolator
    lons_o: numpy.ndarray
        Longitude coordinates of the values usde by the interpolator
    values: numpy.ndarray
        Values usde by the interpolator


    Returns
    --------
    interpolator: function
        Nearest neighbour interpolator function
    """
    # Determine if we are dealing with a regular grid
    if is_regular_grid(lats_o[2:-2, 2:-2], lons_o[2:-2, 2:-2]):
        return _nearest_2D_interpolator_reg(lats_o, lons_o, values)
    else:
        return _nearest_2D_interpolator_arb(lats_o, lons_o, values)


def _nearest_2D_interpolator_reg(lats_o, lons_o, values):
    f = RegularGridInterpolator((np.flipud(lats_o[:, 0]), lons_o[0, :]),
                                np.flipud(values), method='nearest',
                                bounds_error=False)
    return f


def _nearest_2D_interpolator_arb(lats_o, lons_o, values):
    return lambda x: griddata((lats_o.ravel(), lons_o.ravel()), values.ravel(),
                              (x[:, 0], x[:, 1]), 'nearest')


###############################################################################
#                            Bilinear Interpolation                           #
###############################################################################

def bilinear_2D_interpolator(lats_o, lons_o, values):
    """
    Produces a 2D interpolator function using the bilinear interpolation
    method. If the grids are regular grids, uses the
    scipy.interpolate.RegularGridInterpolator,
    otherwise, scipy.intepolate.griddata

    Values can be interpolated from the returned function as follows:

    .. code-block:: python

      f = nearest_2D_interpolator(lat_origin, lon_origin, values_origin)
      interp_values = f(lat_interp, lon_interp)


    Parameters
    -----------
    lats_o: numpy.ndarray
        Latitude coordinates of the values usde by the interpolator
    lons_o: numpy.ndarray
        Longitude coordinates of the values usde by the interpolator
    values: numpy.ndarray
        Values usde by the interpolator


    Returns
    --------
    interpolator: function
        Bilinear interpolator function
    """
    if is_regular_grid(lats_o[2:-2, 2:-2], lons_o[2:-2, 2:-2]):
        return _bilinear_2D_interpolator_reg(lats_o, lons_o, values)
    else:
        return _bilinear_2D_interpolator_arb(lats_o, lons_o, values)


def _bilinear_2D_interpolator_reg(lats_o, lons_o, values):
    f = RegularGridInterpolator((np.flipud(lats_o[:, 0]), lons_o[0, :]),
                                np.flipud(values), method='linear',
                                bounds_error=False)
    return f


def _bilinear_2D_interpolator_arb(lats_o, lons_o, values):
    return lambda x: griddata((lats_o.ravel(), lons_o.ravel()), values.ravel(),
                              (x[:, 0], x[:, 1]), 'linear')


###############################################################################
#                            Bicubic Interpolation                            #
###############################################################################

def bicubic_2D_interpolator(lats_o, lons_o, values):
    """
    Produces a 2D interpolator function using the bicubic interpolation
    method. Uses the scipy.intepolate.griddata method.

    Values can be interpolated from the returned function as follows:

    .. code-block:: python

      f = nearest_2D_interpolator(lat_origin, lon_origin, values_origin)
      interp_values = f(lat_interp, lon_interp)


    Parameters
    -----------
    lats_o: numpy.ndarray
        Latitude coordinates of the values usde by the interpolator
    lons_o: numpy.ndarray
        Longitude coordinates of the values usde by the interpolator
    values: numpy.ndarray
        Values usde by the interpolator


    Returns
    --------
    interpolator: function
        Bicubic interpolator function
    """
    if is_regular_grid(lats_o[2:-2, 2:-2], lons_o[2:-2, 2:-2]):
        return _bicubic_2D_interpolator_reg(lats_o, lons_o, values)
    else:
        return _bicubic_2D_interpolator_arb(lats_o, lons_o, values)


def _bicubic_2D_interpolator_arb(lats_o, lons_o, values):
    return lambda x: griddata((lats_o.ravel(), lons_o.ravel()),
                              values.ravel(), (x[:, 0], x[:, 1]), 'cubic')


def _bicubic_2D_interpolator_reg(lats_o, lons_o, values):
    lat_row = lats_o[1:-1, 1]
    lon_row = lons_o[1, 1:-1]
    I = values

    def K(d):
        d = np.abs(d)
        return np.where(np.logical_and(d >= 0, d <= 1),
                        1.5 * d**3 - 2.5 * d**2 + 1,
                        np.where(np.logical_and(d >= 1, d <= 2),
                                 -0.5 * d**3 + 2.5 * d**2 - 4 * d + 2, 0))

    def interpolator(vect):
        lat = vect[:, 0]
        lon = vect[:, 1]

        # Make sure that we do not hit the limit cases
        R = ((np.searchsorted(lat_row, lat, 'right') - 1) +
             (np.searchsorted(lat_row, lat, 'left') - 1))//2
        C = ((np.searchsorted(lon_row, lon, 'right') - 1) +
             (np.searchsorted(lon_row, lon, 'right') - 1))//2

        diff_lats = np.diff(lat_row)[0]
        diff_lons = np.diff(lon_row)[0]

        r = (lat - lat_row[0]) / diff_lats + 1
        c = (lon - lon_row[0]) / diff_lons + 1

        RI_Rc = I[R, C] * K(c - C) + I[R, C + 1] * K(c - (C + 1)) + \
            I[R, C + 2] * K(c - (C + 2)) + I[R, C + 3] * K(c - (C + 3))
        RI_R1c = I[R + 1, C] * K(c - C) + I[R + 1, C + 1] * K(c - (C + 1)) + \
            I[R + 1, C + 2] * K(c - (C + 2)) + I[R + 1, C + 3] * K(c - (C + 3))
        RI_R2_c = I[R + 2, C] * K(c - C) + I[R + 2, C + 1] * K(c - (C + 1)) +\
            I[R + 2, C + 2] * K(c - (C + 2)) + I[R + 2, C + 3] * K(c - (C + 3))
        RI_R3_c = I[R + 3, C] * K(c - C) + I[R + 3, C + 1] * K(c - (C + 1)) +\
            I[R + 3, C + 2] * K(c - (C + 2)) + I[R + 3, C + 3] * K(c - (C + 3))

        return (RI_Rc * K(r - R) + RI_R1c * K(r - (R + 1)) +
                RI_R2_c * K(r - (R + 2)) + RI_R3_c * K(r - (R + 3)))

    return interpolator



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


import numpy as np
from astropy import units as u
from scipy.optimize import bisect
import scipy.stats as stats

from itur import utils
from itur.models.itu1510 import surface_month_mean_temperature
from itur.models.itu1144 import bilinear_2D_interpolator
from itur.utils import (prepare_input_array, prepare_output_array,
                        load_data_interpolator, prepare_quantity)


class __ITU837():
    """Characteristics of precipitation for propagation modelling

    Available versions include:
    * P.837-6 (02/12) (Superseded)
    * P.837-7 (12/17) (Current version)
    Not-available versions:
    * P.837-1 (08/94) (Superseded)
    * P.837-2 (10/99) (Superseded)
    * P.837-3 (02/01) (Superseded)
    * P.837-4 (04/03) (Superseded)
    * P.837-5 (08/07) (Superseded)

    """
    # This is an abstract class that contains an instance to a version of the
    # ITU-R P.837 recommendation.

    def __init__(self, version=7):
        if version == 7:
            self.instance = _ITU837_7()
        elif version == 6:
            self.instance = _ITU837_6()
#        elif version == 5:
#            self.instance = _ITU837_5()
#        elif version == 4:
#            self.instance = _ITU837_4()
#        elif version == 3:
#            self.instance = _ITU837_3()
#        elif version == 2:
#            self.instance = _ITU837_2()
#        elif version == 1:
#            self.instance = _ITU837_1()
        else:
            raise ValueError(
                'Version {0} is not implemented for the ITU-R P.837 model.'
                .format(version))

        self._Pr6 = {}
        self._Mt = {}
        self._Beta = {}
        self._R001 = {}

    @property
    def __version__(self):
        return self.instance.__version__

    def rainfall_probability(self, lat, lon):
        # Abstract method to compute the rain height
        return self.instance.rainfall_probability(lat, lon)

    def rainfall_rate(self, lat, lon, p):
        # Abstract method to compute the zero isoterm height
        fcn = np.vectorize(self.instance.rainfall_rate, excluded=[0, 1],
                           otypes=[np.ndarray])
        return np.array(fcn(lat, lon, p).tolist())


class _ITU837_7():

    def __init__(self):
        self.__version__ = 7
        self.year = 2017
        self.month = 6
        self.link = 'https://www.itu.int/rec/R-REC-P.837-7-201706-I/en'

        self.months = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
        self._Mt = {}
        self._R001 = {}

    def Mt(self, lat, lon, m):
        if not self._Mt:
            for _m in self.months:
                self._Mt[_m] = load_data_interpolator(
                       '837/v7_lat_mt.npz', '837/v7_lon_mt.npz',
                       '837/v7_mt_month{0:02d}.npz'.format(_m),
                       bilinear_2D_interpolator)

        # In this recommendation the longitude is encoded with format -180 to
        # 180 whereas we always use 0 - 360 encoding
        lon = np.array(lon)
        lon[lon > 180] = lon[lon > 180] - 360
        return self._Mt[m](
            np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    def R001(self, lat, lon):
        if not self._R001:
            self._R001 = load_data_interpolator(
                       '837/v7_lat_r001.npz', '837/v7_lon_r001.npz',
                       '837/v7_r001.npz', bilinear_2D_interpolator)

        # In this recommendation the longitude is encoded with format -180 to
        # 180 whereas we always use 0 - 360 encoding
        lon = np.array(lon)
        lon[lon > 180] = lon[lon > 180] - 360
        return self._R001(
            np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    def rainfall_probability(self, lat_d, lon_d):
        """

        """
        lat_f = lat_d.flatten()
        lon_f = lon_d.flatten()

        Nii = np.array([[31, 28.25, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]])

        # Step 2: For each month, determine the monthly mean surface
        # temperature
        Tii = surface_month_mean_temperature(lat_f, lon_f, self.months).value.T

        # Step 3: For each month, determine the monthly mean total rainfall
        MTii = np.array([self.Mt(lat_f, lon_f, m) for m in self.months]).T

        # Step 4: For each month, determine the monthly mean total rainfall
        tii = Tii - 273.15

        # Step 5: For each month number, calculate rii
        rii = np.where(tii >= 0, 0.5874 * np.exp(0.0883 * tii), 0.5874)  # Eq.1

        # Step 6a For each month number, calculate the probability of rain:
        P0ii = 100 * MTii / (24 * Nii * rii)  # Eq. 2

        # Step 7b:
        rii = np.where(P0ii > 70, 100/70. * MTii / (24 * Nii), rii)
        P0ii = np.where(P0ii > 70, 70, P0ii)

        # Step 7: Calculate the annual probability of rain, P0anual
        P0anual = np.sum(Nii * P0ii, axis=-1) / 365.25  # Eq. 3

        return P0anual.reshape(lat_d.shape)

    def rainfall_rate(self, lat_d, lon_d, p):
        """
        """
        if p == 0.01:
            return self.R001(lat_d, lon_d)

        lat_f = lat_d.flatten()
        lon_f = lon_d.flatten()

        Nii = np.array([[31, 28.25, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]])

        # Step 2: For each month, determine the monthly mean surface
        # temperature
        Tii = surface_month_mean_temperature(lat_f, lon_f, self.months).value.T

        # Step 3: For each month, determine the monthly mean total rainfall
        MTii = np.array([self.Mt(lat_f, lon_f, m) for m in self.months]).T

        # Step 4: For each month, determine the monthly mean total rainfall
        tii = Tii - 273.15

        # Step 5: For each month number, calculate rii
        rii = np.where(tii >= 0, 0.5874 * np.exp(0.0883 * tii), 0.5874)

        # Step 6a For each month number, calculate the probability of rain:
        P0ii = 100 * MTii / (24 * Nii * rii)

        # Step 7b:
        rii = np.where(P0ii > 70, 100/70. * MTii / (24 * Nii), rii)
        P0ii = np.where(P0ii > 70, 70, P0ii)

        # Step 7: Calculate the annual probability of rain, P0anual
        P0anual = np.sum(Nii * P0ii, axis=-1) / 365.25

        # Step 8: Compute the rainfall rate exceeded for p
        def _ret_fcn(P0):
            if p > P0:
                return 0
            else:
                # Use a bisection method to determine
                def f_Rp(Rref):
                    P_r_ge_Rii = P0ii * stats.norm.sf(
                            (np.log(Rref) + 0.7938 - np.log(rii)) / 1.26)
                    P_r_ge_R = np.sum(Nii * P_r_ge_Rii) / 365.25
                    return 100 * (P_r_ge_R / p - 1)

                return bisect(f_Rp, 1e-10, 1000, xtol=1e-5)

        fcn = np.vectorize(_ret_fcn)
        return fcn(P0anual).reshape(lat_d.shape)


class _ITU837_6():

    def __init__(self):
        self.__version__ = 6
        self.year = 2012
        self.month = 2
        self.link = 'https://www.itu.int/rec/R-REC-P.837-6-201202-I/en'

        self._Pr6 = {}
        self._Mt = {}
        self._Beta = {}

    def Pr6(self, lat, lon):
        if not self._Pr6:
            self._Pr6 = load_data_interpolator(
                       '837/esarain_lat_v5.npz', '837/esarain_lon_v5.npz',
                       '837/esarain_pr6_v5.npz', bilinear_2D_interpolator,
                       flip_ud=False)


        return self._Pr6(
            np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    def Mt(self, lat, lon):
        if not self._Mt:
            self._Mt = load_data_interpolator(
                       '837/esarain_lat_v5.npz', '837/esarain_lon_v5.npz',
                       '837/esarain_mt_v5.npz', bilinear_2D_interpolator,
                       flip_ud=False)

        return self._Mt(
            np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    def Beta(self, lat, lon):
        if not self._Beta:
            self._Beta = load_data_interpolator(
                       '837/esarain_lat_v5.npz', '837/esarain_lon_v5.npz',
                       '837/esarain_beta_v5.npz', bilinear_2D_interpolator,
                       flip_ud=False)

        return self._Beta(
            np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    def rainfall_probability(self, lat_d, lon_d):
        """

        """
        Pr6 = self.Pr6(lat_d, lon_d)
        Mt = self.Mt(lat_d, lon_d)
        Beta = self.Beta(lat_d, lon_d)

        # Step 3: Convert MT and β to Mc and Ms as follows:
        Ms = (1 - Beta) * Mt

        # Step 4: Derive the percentage propability of rain in an average year,
        # P0:
        P0 = Pr6 * (1 - np.exp(-0.0079 * (Ms / Pr6)))  # Eq. 1

        return P0

    def rainfall_rate(self, lat_d, lon_d, p):
        """
        """
        Pr6 = self.Pr6(lat_d, lon_d)
        Mt = self.Mt(lat_d, lon_d)
        Beta = self.Beta(lat_d, lon_d)

        # Step 3: Convert MT and β to Mc and Ms as follows:
        Mc = Beta * Mt
        Ms = (1 - Beta) * Mt

        # Step 4: Derive the percentage propability of rain in an average year,
        # P0:
        P0 = np.where(Pr6 > 0,
                      Pr6 * (1 - np.exp(-0.0079 * (Ms / Pr6))),
                      0)  # Eq. 1

        # Step 5: Derive the rainfall rate, Rp, exceeded for p% of the average
        # year, where p <= P0, from:
        def computeRp(P0, Mc, Ms):
            a = 1.09                        # Eq. 2d
            b = (Mc + Ms) / (21797 * P0)    # Eq. 2e
            c = 26.02 * b                   # Eq. 2f

            A = a * b                       # Eq. 2a
            B = a + c * np.log(p / P0)      # Eq. 2b
            C = np.log(p / P0)              # Eq. 2c

            Rp = (-B + np.sqrt(B**2 - 4 * A * C)) / (2 * A)  # Eq. 2

            return Rp

        # The value of Rp can only be computed for those values where p > P0
        Rp = np.where(np.isnan(P0) | (p > P0), 0, computeRp(P0, Mc, Ms))
        return Rp


__model = __ITU837()


def change_version(new_version):
    """
    Change the version of the ITU-R P.837 recommendation currently being used.


    This function changes the model used for the ITU-R P.837 recommendation
    to a different version.

    Parameters
    ----------
    new_version : int
        Number of the version to use.
        Valid values are:
          *  7: Activates recommendation ITU-R P.837-7 (12/17) (Current version)
          *  6: Activates recommendation ITU-R P.837-6 (02/12) (Superseded)

    """
    global __model
    __model = __ITU837(new_version)


def get_version():
    """
    Obtain the version of the ITU-R P.837 recommendation currently being used.

    Returns
    -------
    version: int
        Version currently being used.
    """
    return __model.__version__


def rainfall_probability(lat, lon):
    """
    Compute the percentage probability of rain in an average year, P0,  at a
    given location.


    Parameters
    ----------
    lat : number, sequence, or numpy.ndarray
        Latitudes of the receiver points
    lon : number, sequence, or numpy.ndarray
        Longitudes of the receiver points


    Returns
    -------
    P0: numpy.ndarray
        Percentage probability of rain in an average year (%)


    References
    ----------
    [1] Characteristics of precipitation for propagation modelling
    https://www.itu.int/rec/R-REC-P.837/en
    """
    type_output = type(lat)
    lat = prepare_input_array(lat)
    lon = prepare_input_array(lon)
    lon = np.mod(lon, 360)
    val = __model.rainfall_probability(lat, lon)
    return prepare_output_array(val, type_output) * u.pct


def rainfall_rate(lat, lon, p):
    """
    Compute the rainfall rate exceeded for p% of the average year at a
    given location.


    Parameters
    ----------
    lat : number, sequence, or numpy.ndarray
        Latitudes of the receiver points
    lon : number, sequence, or numpy.ndarray
        Longitudes of the receiver points
    p : number
        Percentage of time exceeded for p% of the average year


    Returns
    -------
    R001: numpy.ndarray
        Rainfall rate exceeded for p% of the average year


    References
    ----------
    [1] Characteristics of precipitation for propagation modelling
    https://www.itu.int/rec/R-REC-P.837/en
    """
    type_output = type(lat)
    lat = prepare_input_array(lat)
    lon = prepare_input_array(lon)
    lon = np.mod(lon, 360)
    val = __model.rainfall_rate(lat, lon, p)
    return prepare_output_array(val, type_output) * u.mm / u.hr


def unavailability_from_rainfall_rate(lat, lon, R):
    """Compute the percentage of time of the average year that a given rainfall
    rate (R) is exceeded at a given location

    This method calls successively to `rainfall_rate` (sing bisection) with
    different values of p.


    Parameters
    ----------
    lat : number, sequence, or numpy.ndarray
        Latitudes of the receiver points
    lon : number, sequence, or numpy.ndarray
        Longitudes of the receiver points
    R : number, sequence, or numpy.ndarray
        Rainfall rate (mm/h)


    Returns
    -------
    p: numpy.ndarray
        Rainfall rate exceeded for p% of the average year


    References
    ----------
    [1] Characteristics of precipitation for propagation modelling
    https://www.itu.int/rec/R-REC-P.837/en
    """
    lat = prepare_input_array(lat)
    lon = prepare_input_array(lon)
    lon = np.mod(lon, 360)
    R = prepare_quantity(R, u.mm / u.hr, 'Rain rate')

    #TODO: Cehck for bound on R (between 0 and 200 mm/hr?)

    def fcn(x):
        return (rainfall_rate(lat, lon, x).value - R - 1e-6)

    return bisect(fcn, 1e-5, 100, maxiter=50)
